"""Repl server cli."""
import logging
import shutil

import click
from prompt_toolkit import PromptSession, print_formatted_text
from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import clear
from prompt_toolkit.shortcuts.progress_bar import formatters
from prompt_toolkit.styles import Style


_logger = logging.getLogger(__name__)

TITLE = r"""
__________                          .______.                    _________
\______   \___.__. _____   ____   __| _/\_ |__  __ __  ______  /   _____/ ______________  __ ___________
 |     ___<   |  |/     \ /  _ \ / __ |  | __ \|  |  \/  ___/  \_____  \_/ __ \_  __ \  \/ // __ \_  __ \\
 |    |    \___  |  Y Y  (  <_> ) /_/ |  | \_\ \  |  /\___ \   /        \  ___/|  | \/\   /\  ___/|  | \/
 |____|    / ____|__|_|  /\____/\____ |  |___  /____//____  > /_______  /\___  >__|    \_/  \___  >__|
           \/          \/            \/      \/           \/          \/     \/                 \/
"""


SMALL_TITLE = "Pymodbus server..."
BOTTOM_TOOLBAR = HTML(
    '(MODBUS SERVER) <b><style bg="ansired">Press Ctrl+C or '
    'type "exit" to quit</style></b> Type "help" '
    "for list of available commands"
)
COMMAND_ARGS = ["response_type", "error_code", "delay_by", "clear_after", "data_len"]
RESPONSE_TYPES = ["normal", "error", "delayed", "empty", "stray"]
COMMANDS = {
    "manipulator": {
        "response_type": None,
        "error_code": None,
        "delay_by": None,
        "clear_after": None,
    },
    "exit": None,
    "help": None,
    "clear": None,
}
USAGE = (
    "manipulator response_type=|normal|error|delayed|empty|stray \n"
    "\tAdditional parameters\n"
    "\t\terror_code=&lt;int&gt; \n\t\tdelay_by=&lt;in seconds&gt; \n\t\t"
    "clear_after=&lt;clear after n messages int&gt;"
    "\n\t\tdata_len=&lt;length of stray data (int)&gt;\n"
    "\n\tExample usage: \n\t"
    "1. Send error response 3 for 4 requests\n\t"
    "   <ansiblue>manipulator response_type=error error_code=3 clear_after=4</ansiblue>\n\t"
    "2. Delay outgoing response by 5 seconds indefinitely\n\t"
    "   <ansiblue>manipulator response_type=delayed delay_by=5</ansiblue>\n\t"
    "3. Send empty response\n\t"
    "   <ansiblue>manipulator response_type=empty</ansiblue>\n\t"
    "4. Send stray response of length 12 and revert to normal after 2 responses\n\t"
    "   <ansiblue>manipulator response_type=stray data_len=11 clear_after=2</ansiblue>\n\t"
    "5. To disable response manipulation\n\t"
    "   <ansiblue>manipulator response_type=normal</ansiblue>"
)
COMMAND_HELPS = {
    "manipulator": f"Manipulate response from server.\nUsage: {USAGE}",
    "clear": "Clears screen",
}


STYLE = Style.from_dict({"": "cyan"})
CUSTOM_FORMATTERS = [
    formatters.Label(suffix=": "),
    formatters.Bar(start="|", end="|", sym_a="#", sym_b="#", sym_c="-"),
    formatters.Text(" "),
    formatters.Text(" "),
    formatters.TimeElapsed(),
    formatters.Text("  "),
]


def info(message):
    """Show info."""
    if not isinstance(message, str):
        message = str(message)
    click.secho(message, fg="green")


def warning(message):
    """Show warning."""
    click.secho(str(message), fg="yellow")


def error(message):
    """Show error."""
    click.secho(str(message), fg="red")


def get_terminal_width():
    """Get terminal width."""
    return shutil.get_terminal_size()[0]


def print_help():
    """Print help."""
    print_formatted_text(HTML("<u>Available commands:</u>"))
    for cmd, hlp in sorted(COMMAND_HELPS.items()):
        print_formatted_text(
            HTML("<skyblue>{cmd:45s}</skyblue><seagreen>{hlp:100s}</seagreen>")
        )


async def interactive_shell(server):  # pylint: disable=too-complex
    """Run CLI interactive shell."""
    col = get_terminal_width()
    max_len = max(  # pylint: disable=consider-using-generator
        [len(t) for t in TITLE.split("\n")]
    )
    if col > max_len:
        info(TITLE)
    else:
        print_formatted_text(
            HTML(f'<u><b><style color="green">{SMALL_TITLE}</style></b></u>')
        )
    info("")
    completer = NestedCompleter.from_nested_dict(COMMANDS)
    session = PromptSession(
        "SERVER > ", completer=completer, bottom_toolbar=BOTTOM_TOOLBAR
    )

    # Run echo loop. Read text from stdin, and reply it back.
    while True:  # pylint: disable=too-many-nested-blocks
        try:
            invalid_command = False
            result = await session.prompt_async()
            if result == "exit":
                await server.web_app.shutdown()
                break
            if result == "help":
                print_help()
                continue
            if result == "clear":
                clear()
                continue
            if command := result.split():
                if command[0] not in COMMANDS:
                    invalid_command = True
                if invalid_command:
                    warning(f"Invalid command or invalid usage of command - {command}")
                    continue
                if len(command) == 1:
                    warning(f'Usage: "{USAGE}"')
                else:
                    args = command[1:]
                    skip_next = False
                    val_dict = {}
                    for index, arg in enumerate(args):
                        if skip_next:
                            skip_next = False
                            continue
                        if "=" in arg:
                            arg, value = arg.split("=")
                        elif arg in COMMAND_ARGS:
                            try:
                                value = args[index + 1]
                                skip_next = True
                            except IndexError:
                                error(f"Missing value for argument - {arg}")
                                warning('Usage: "{USAGE}"')
                                break
                        valid = True
                        if arg == "response_type":
                            if value not in RESPONSE_TYPES:
                                warning(f"Invalid response type request - {value}")
                                warning(f"Choose from {RESPONSE_TYPES}")
                                valid = False
                        elif arg in {  # pylint: disable=confusing-consecutive-elif
                            "error_code",
                            "delay_by",
                            "clear_after",
                            "data_len",
                        }:
                            try:
                                value = int(value)
                            except ValueError:
                                warning(f"Expected integer value for {arg}")
                                valid = False

                        if valid:
                            val_dict[arg] = value
                    if val_dict:
                        server.update_manipulator_config(val_dict)
                        # server.manipulator_config = val_dict
                # result = await run_command(tester, *command)

        except (EOFError, KeyboardInterrupt):
            return


async def main(server):
    """Run main."""
    # with patch_stdout():
    try:
        await interactive_shell(server)
    finally:
        pass
    warning("Bye Bye!!!")


async def run_repl(server):
    """Run repl server."""
    await main(server)



"""Binary framer."""
# pylint: disable=missing-type-doc
import logging
import struct

from pymodbus.exceptions import ModbusIOException
from pymodbus.framer import BYTE_ORDER, FRAME_HEADER, ModbusFramer
from pymodbus.utilities import checkCRC, computeCRC


# --------------------------------------------------------------------------- #
# Logging
# --------------------------------------------------------------------------- #
_logger = logging.getLogger(__name__)

BINARY_FRAME_HEADER = BYTE_ORDER + FRAME_HEADER

# --------------------------------------------------------------------------- #
# Modbus Binary Message
# --------------------------------------------------------------------------- #


class ModbusBinaryFramer(ModbusFramer):
    """Modbus Binary Frame Controller.

        [ Start ][Address ][ Function ][ Data ][ CRC ][ End ]
          1b        1b         1b         Nb     2b     1b

        * data can be 0 - 2x252 chars
        * end is   "}"
        * start is "{"

    The idea here is that we implement the RTU protocol, however,
    instead of using timing for message delimiting, we use start
    and end of message characters (in this case { and }). Basically,
    this is a binary framer.

    The only case we have to watch out for is when a message contains
    the { or } characters.  If we encounter these characters, we
    simply duplicate them.  Hopefully we will not encounter those
    characters that often and will save a little bit of bandwitch
    without a real-time system.

    Protocol defined by jamod.sourceforge.net.
    """

    method = "binary"

    def __init__(self, decoder, client=None):
        """Initialize a new instance of the framer.

        :param decoder: The decoder implementation to use
        """
        self._buffer = b""
        self._header = {"crc": 0x0000, "len": 0, "uid": 0x00}
        self._hsize = 0x01
        self._start = b"\x7b"  # {
        self._end = b"\x7d"  # }
        self._repeat = [b"}"[0], b"{"[0]]  # python3 hack
        self.decoder = decoder
        self.client = client

    # ----------------------------------------------------------------------- #
    # Private Helper Functions
    # ----------------------------------------------------------------------- #
    def decode_data(self, data):
        """Decode data."""
        if len(data) > self._hsize:
            uid = struct.unpack(">B", data[1:2])[0]
            fcode = struct.unpack(">B", data[2:3])[0]
            return {"unit": uid, "fcode": fcode}
        return {}

    def checkFrame(self):
        """Check and decode the next frame.

        :returns: True if we are successful, False otherwise
        """
        start = self._buffer.find(self._start)
        if start == -1:
            return False
        if start > 0:  # go ahead and skip old bad data
            self._buffer = self._buffer[start:]

        if (end := self._buffer.find(self._end)) != -1:
            self._header["len"] = end
            self._header["uid"] = struct.unpack(">B", self._buffer[1:2])[0]
            self._header["crc"] = struct.unpack(">H", self._buffer[end - 2 : end])[0]
            data = self._buffer[start + 1 : end - 2]
            return checkCRC(data, self._header["crc"])
        return False

    def advanceFrame(self):
        """Skip over the current framed message.

        This allows us to skip over the current message after we have processed
        it or determined that it contains an error. It also has to reset the
        current frame header handle
        """
        self._buffer = self._buffer[self._header["len"] + 2 :]
        self._header = {"crc": 0x0000, "len": 0, "uid": 0x00}

    def isFrameReady(self):
        """Check if we should continue decode logic.

        This is meant to be used in a while loop in the decoding phase to let
        the decoder know that there is still data in the buffer.

        :returns: True if ready, False otherwise
        """
        return len(self._buffer) > 1

    def addToFrame(self, message):
        """Add the next message to the frame buffer.

        This should be used before the decoding while loop to add the received
        data to the buffer handle.

        :param message: The most recent packet
        """
        self._buffer += message

    def getFrame(self):
        """Get the next frame from the buffer.

        :returns: The frame data or ""
        """
        start = self._hsize + 1
        end = self._header["len"] - 2
        buffer = self._buffer[start:end]
        if end > 0:
            return buffer
        return b""

    def populateResult(self, result):
        """Populate the modbus result header.

        The serial packets do not have any header information
        that is copied.

        :param result: The response packet
        """
        result.unit_id = self._header["uid"]

    # ----------------------------------------------------------------------- #
    # Public Member Functions
    # ----------------------------------------------------------------------- #
    def processIncomingPacket(
        self, data, callback, unit, **kwargs
    ):  # pylint: disable=arguments-differ
        """Process new packet pattern.

        This takes in a new request packet, adds it to the current
        packet stream, and performs framing on it. That is, checks
        for complete messages, and once found, will process all that
        exist.  This handles the case when we read N + 1 or 1 // N
        messages at a time instead of 1.

        The processed and decoded messages are pushed to the callback
        function to process and send.

        :param data: The new packet data
        :param callback: The function to send results to
        :param unit: Process if unit id matches, ignore otherwise (could be a
               list of unit ids (server) or single unit id(client/server)
        :param kwargs:
        :raises ModbusIOException:
        """
        self.addToFrame(data)
        if not isinstance(unit, (list, tuple)):
            unit = [unit]
        single = kwargs.get("single", False)
        while self.isFrameReady():
            if self.checkFrame():
                if self._validate_unit_id(unit, single):
                    if (result := self.decoder.decode(self.getFrame())) is None:
                        raise ModbusIOException("Unable to decode response")
                    self.populateResult(result)
                    self.advanceFrame()
                    callback(result)  # defer or push to a thread?
                else:
                    header_txt = self._header["uid"]
                    txt = f"Not a valid unit id - {header_txt}, ignoring!!"
                    _logger.debug(txt)
                    self.resetFrame()
                    break

            else:
                _logger.debug("Frame check failed, ignoring!!")
                self.resetFrame()
                break

    def buildPacket(self, message):
        """Create a ready to send modbus packet.

        :param message: The request/response to send
        :returns: The encoded packet
        """
        data = self._preflight(message.encode())
        packet = (
            struct.pack(BINARY_FRAME_HEADER, message.unit_id, message.function_code)
            + data
        )
        packet += struct.pack(">H", computeCRC(packet))
        packet = self._start + packet + self._end
        return packet

    def _preflight(self, data):
        """Do preflight buffer test.

        This basically scans the buffer for start and end
        tags and if found, escapes them.

        :param data: The message to escape
        :returns: the escaped packet
        """
        array = bytearray()
        for item in data:
            if item in self._repeat:
                array.append(item)
            array.append(item)
        return bytes(array)

    def resetFrame(self):  # pylint: disable=invalid-name
        """Reset the entire message frame.

        This allows us to skip ovver errors that may be in the stream.
        It is hard to know if we are simply out of sync or if there is
        an error in the stream as we have no way to check the start or
        end of the message (python just doesn"t have the resolution to
        check for millisecond delays).
        """
        self._buffer = b""
        self._header = {"crc": 0x0000, "len": 0, "uid": 0x00}


# __END__



"""Sunspec client."""
import logging

from twisted.internet.defer import Deferred  # pylint: disable=import-error

from pymodbus.client.sync import ModbusTcpClient
from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadDecoder


# --------------------------------------------------------------------------- #
# Logging
# --------------------------------------------------------------------------- #
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)


# --------------------------------------------------------------------------- #
# Sunspec Common Constants
# --------------------------------------------------------------------------- #
class SunspecDefaultValue:
    """A collection of constants to indicate if a value is not implemented."""

    Signed16 = 0x8000
    Unsigned16 = 0xFFFF
    Accumulator16 = 0x0000
    Scale = 0x8000
    Signed32 = 0x80000000
    Float32 = 0x7FC00000
    Unsigned32 = 0xFFFFFFFF
    Accumulator32 = 0x00000000
    Signed64 = 0x8000000000000000
    Unsigned64 = 0xFFFFFFFFFFFFFFFF
    Accumulator64 = 0x0000000000000000
    String = "\x00"


class SunspecStatus:
    """Indicators of the current status of a sunspec device"""

    Normal = 0x00000000
    Error = 0xFFFFFFFE
    Unknown = 0xFFFFFFFF


class SunspecIdentifier:
    """Assigned identifiers that are pre-assigned by the sunspec protocol."""

    Sunspec = 0x53756E53


class SunspecModel:
    """Assigned device indentifiers that are pre-assigned by the sunspec protocol."""

    # ---------------------------------------------
    # 0xx Common Models
    # ---------------------------------------------
    CommonBlock = 1
    AggregatorBlock = 2

    # ---------------------------------------------
    # 1xx Inverter Models
    # ---------------------------------------------
    SinglePhaseIntegerInverter = 101
    SplitPhaseIntegerInverter = 102
    ThreePhaseIntegerInverter = 103
    SinglePhaseFloatsInverter = 103
    SplitPhaseFloatsInverter = 102
    ThreePhaseFloatsInverter = 103

    # ---------------------------------------------
    # 2xx Meter Models
    # ---------------------------------------------
    SinglePhaseMeter = 201
    SplitPhaseMeter = 201
    WyeConnectMeter = 201
    DeltaConnectMeter = 201

    # ---------------------------------------------
    # 3xx Environmental Models
    # ---------------------------------------------
    BaseMeteorological = 301
    Irradiance = 302
    BackOfModuleTemperature = 303
    Inclinometer = 304
    Location = 305
    ReferencePoint = 306
    BaseMeteorological = 307
    MiniMeteorological = 308

    # ---------------------------------------------
    # 4xx String Combiner Models
    # ---------------------------------------------
    BasicStringCombiner = 401
    AdvancedStringCombiner = 402

    # ---------------------------------------------
    # 5xx Panel Models
    # ---------------------------------------------
    PanelFloat = 501
    PanelInteger = 502

    # ---------------------------------------------
    # 641xx outback_ Blocks
    # ---------------------------------------------
    outback_device_identifier = 64110
    outback_charge_controller = 64111
    outback_fm_charge_controller = 64112
    outback_fx_inv_realtime = 64113
    outback_fx_inv_conf = 64114
    outback_split_phase_rad_inv = 64115
    outback_radian_inv_conf = 64116
    outback_single_phase_rad_inv_rt = 64117
    outback_flexnet_dc_realtime = 64118
    outback_flexnet_dc_conf = 64119
    outback_system_control = 64120

    # ---------------------------------------------
    # 64xxx Vendor Extension Block
    # ---------------------------------------------
    EndOfSunSpecMap = 65535

    @classmethod
    def lookup(cls, code):
        """Return the device model name for that identifier

        :param code: The device code to lookup
        :returns: The device model name, or None if none available
        """
        values = {(v, k) for k, v in cls.__dict__.iteritems() if not callable(v)}
        return values.get(code, None)


class SunspecOffsets:
    """Well known offsets that are used throughout the sunspec protocol"""

    CommonBlock = 40000
    CommonBlockLength = 69
    AlternateCommonBlock = 50000


# --------------------------------------------------------------------------- #
# Common Functions
# --------------------------------------------------------------------------- #
def defer_or_apply(func):
    """Apply an adapter method.

    to a result regardless if it is a deferred
    or a concrete response.

    :param func: The function to decorate
    """

    def closure(future, adapt):
        if isinstance(future, Deferred):
            defer = Deferred()
            future.addCallback(lambda r: defer.callback(adapt(r)))
            return defer
        return adapt(future)

    return closure


def create_sunspec_sync_client(host):
    """Create a sunspec client.

    :param host: The host to connect to
    :returns: an initialized SunspecClient
    """
    modbus = ModbusTcpClient(host)
    modbus.connect()
    client = SunspecClient(modbus)
    client.initialize()
    return client


# --------------------------------------------------------------------------- #
# Sunspec Client
# --------------------------------------------------------------------------- #
class SunspecDecoder(BinaryPayloadDecoder):
    """A decoder that deals correctly with the sunspec binary format."""

    def __init__(self, payload, byteorder):
        """Initialize a new instance of the SunspecDecoder

        .. note:: This is always set to big endian byte order
        as specified in the protocol.
        """
        my_byteorder = Endian.Big
        BinaryPayloadDecoder.__init__(self, payload, my_byteorder)

    def decode_string(self, size=1):
        """Decode a string from the buffer

        :param size: The size of the string to decode
        """
        self._pointer += size
        string = self._payload[self._pointer - size : self._pointer]
        return string.split(SunspecDefaultValue.String)[0]


class SunspecClient:
    """SunSpec client."""

    def __init__(self, client):
        """Initialize a new instance of the client

        :param client: The modbus client to use
        """
        self.client = client
        self.offset = SunspecOffsets.CommonBlock

    def initialize(self):
        """Initialize the underlying client values

        :returns: True if successful, false otherwise
        """
        decoder = self.get_device_block(self.offset, 2)
        if decoder.decode_32bit_uint() == SunspecIdentifier.Sunspec:
            return True
        self.offset = SunspecOffsets.AlternateCommonBlock
        decoder = self.get_device_block(self.offset, 2)
        return decoder.decode_32bit_uint() == SunspecIdentifier.Sunspec

    def get_common_block(self):
        """Read and return the sunspec common information block.

        :returns: A dictionary of the common block information
        """
        length = SunspecOffsets.CommonBlockLength
        decoder = self.get_device_block(self.offset, length)
        return {
            "SunSpec_ID": decoder.decode_32bit_uint(),
            "SunSpec_DID": decoder.decode_16bit_uint(),
            "SunSpec_Length": decoder.decode_16bit_uint(),
            "Manufacturer": decoder.decode_string(size=32),
            "Model": decoder.decode_string(size=32),
            "Options": decoder.decode_string(size=16),
            "Version": decoder.decode_string(size=16),
            "SerialNumber": decoder.decode_string(size=32),
            "DeviceAddress": decoder.decode_16bit_uint(),
            "Next_DID": decoder.decode_16bit_uint(),
            "Next_DID_Length": decoder.decode_16bit_uint(),
        }

    def get_device_block(self, offset, size):
        """Retrieve the next device block

        .. note:: We will read 2 more registers so that we have
        the information for the next block.

        :param offset: The offset to start reading at
        :param size: The size of the offset to read
        :returns: An initialized decoder for that result
        """
        txt = f"reading device block[{offset}..{offset + size}]"
        _logger.debug(txt)
        response = self.client.read_holding_registers(offset, size + 2)
        return SunspecDecoder.fromRegisters(response.registers)

    def get_all_device_blocks(self):
        """Retrieve all the available blocks in the supplied sunspec device.

        .. note:: Since we do not know how to decode the available
        blocks, this returns a list of dictionaries of the form:

            decoder: the-binary-decoder,
            model:   the-model-identifier (name)

        :returns: A list of the available blocks
        """
        blocks = []
        offset = self.offset + 2
        model = SunspecModel.CommonBlock
        while model != SunspecModel.EndOfSunSpecMap:
            decoder = self.get_device_block(offset, 2)
            model = decoder.decode_16bit_uint()
            length = decoder.decode_16bit_uint()
            blocks.append(
                {
                    "model": model,
                    "name": SunspecModel.lookup(model),
                    "length": length,
                    "offset": offset + length + 2,
                }
            )
            offset += length + 2
        return blocks


# ------------------------------------------------------------
# A quick test runner
# ------------------------------------------------------------
if __name__ == "__main__":
    client = create_sunspec_sync_client("YOUR.HOST.GOES.HERE")

    # print out all the device common block
    common = client.get_common_block()
    for key, value in common.iteritems():
        if key == "SunSpec_DID":
            value = SunspecModel.lookup(value)
        print("{:<20}: {}".format(key, value))

    # print out all the available device blocks
    blocks = client.get_all_device_blocks()
    for block in blocks:
        print(block)

    client.client.close()



"""Test interfaces."""
import unittest

from pymodbus.exceptions import NotImplementedException
from pymodbus.interfaces import (
    IModbusDecoder,
    IModbusFramer,
    IModbusSlaveContext,
    IPayloadBuilder,
    Singleton,
)


class _SingleInstance(Singleton):  # pylint: disable=too-few-public-methods
    """Single instance."""


class ModbusInterfaceTestsTest(unittest.TestCase):
    """Unittest for the pymodbus.interfaces module."""

    def setUp(self):
        """Initialize the test environment"""

    def tearDown(self):
        """Clean up the test environment"""

    def test_singleton_interface(self):
        """Test that the singleton interface works"""
        first = _SingleInstance()
        second = _SingleInstance()
        self.assertEqual(first, second)

    def test_imodbusdecoder(self):
        """Test that the base class isn"t implemented"""
        x_base = None
        instance = IModbusDecoder()
        self.assertRaises(NotImplementedException, lambda: instance.decode(x_base))
        self.assertRaises(
            NotImplementedException, lambda: instance.lookupPduClass(x_base)
        )
        self.assertRaises(NotImplementedException, lambda: instance.register(x_base))

    def test_modbus_framer_interface(self):
        """Test that the base class isn"t implemented"""
        x_base = None
        instance = IModbusFramer()
        self.assertRaises(NotImplementedException, instance.checkFrame)
        self.assertRaises(NotImplementedException, instance.advanceFrame)
        self.assertRaises(NotImplementedException, instance.isFrameReady)
        self.assertRaises(NotImplementedException, instance.getFrame)
        self.assertRaises(NotImplementedException, lambda: instance.addToFrame(x_base))
        self.assertRaises(
            NotImplementedException, lambda: instance.populateResult(x_base)
        )
        self.assertRaises(
            NotImplementedException,
            lambda: instance.processIncomingPacket(x_base, x_base),
        )
        self.assertRaises(NotImplementedException, lambda: instance.buildPacket(x_base))

    def test_modbus_slave_context_interface(self):
        """Test that the base class isn"t implemented"""
        x_base = None
        instance = IModbusSlaveContext()
        self.assertRaises(NotImplementedException, instance.reset)
        self.assertRaises(
            NotImplementedException, lambda: instance.validate(x_base, x_base, x_base)
        )
        self.assertRaises(
            NotImplementedException, lambda: instance.getValues(x_base, x_base, x_base)
        )
        self.assertRaises(
            NotImplementedException, lambda: instance.setValues(x_base, x_base, x_base)
        )

    def test_modbus_payload_builder_interface(self):
        """Test that the base class isn"t implemented"""
        instance = IPayloadBuilder()
        self.assertRaises(
            NotImplementedException,
            lambda: instance.build(),  # pylint: disable=unnecessary-lambda
        )



#!/usr/bin/env python3
"""Simple Modbus TCP over TLS client.

This is a simple example of writing a modbus TCP over TLS client that uses
Python builtin module ssl - TLS/SSL wrapper for socket objects for the TLS
feature.
"""
# -------------------------------------------------------------------------- #
# import necessary libraries
# -------------------------------------------------------------------------- #
import ssl

from pymodbus.client import ModbusTlsClient


# -------------------------------------------------------------------------- #
# the TLS detail security can be set in SSLContext which is the context here
# -------------------------------------------------------------------------- #
sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
sslctx.verify_mode = ssl.CERT_REQUIRED
sslctx.check_hostname = True

# Prepare client"s certificate which the server requires for TLS full handshake
# sslctx.load_cert_chain(certfile="client.crt", keyfile="client.key",
#                        password="pwd")

# -------------------------------------------------------------------------- #
# pass SSLContext which is the context here to ModbusTcpClient()
# -------------------------------------------------------------------------- #
client = ModbusTlsClient("test.host.com", 8020, sslctx=sslctx)
client.connect()

result = client.read_coils(1, 8)
print(result.bits)

client.close()



"""Handle the version information here.

you should only have to change the version tuple.
"""


class Version:
    """Manage version."""

    def __init__(self, package, major, minor, micro, pre=None):
        """Initialize.

        :param package: Name of the package that this is a version of.
        :param major: The major version number.
        :param minor: The minor version number.
        :param micro: The micro version number.
        :param pre: The pre release tag
        """
        self.package = package
        self.major = major
        self.minor = minor
        self.micro = micro
        self.pre = pre

    def short(self):
        """Return a string in canonical short version format: <major>.<minor>.<micro>.<pre>."""
        pre = ""
        if self.pre:
            pre = f".{self.pre}"
        return f"{self.major}.{self.minor}.{self.micro}{pre}"

    def __str__(self):
        """Return a string representation of the object.

        :returns: A string representation of this object
        """
        return f"[{self.package}, version {self.short()}]"


version = Version("pymodbus", 3, 0, 0, "rc1")
version.__name__ = (  # fix epydoc error # pylint: disable=attribute-defined-outside-init
    "pymodbus"
)


# --------------------------------------------------------------------------- #
# Exported symbols
# --------------------------------------------------------------------------- #

__all__ = [
    "version"
]



#!/usr/bin/env python3
"""Epydoc API Runner.

Using pkg_resources, we attempt to see if epydoc is installed,
if so, we use its cli program to compile the documents
"""
import os
import shutil
import sys
import traceback

import pkg_resources

try:
    pkg_resources.require("epydoc")

    from epydoc.cli import cli  # pylint: disable=import-error

    sys.argv = """epydoc.py pymodbus
        --html --simple-term --quiet
        --include-log
        --graph=all
        --docformat=plaintext
        --debug
        --exclude=._
        --exclude=tests
        --output=html/
    """.split()
    #  bugs in trunk for --docformat=restructuredtext

    if not os.path.exists("./html"):
        os.mkdir("./html")

    print("Building Epydoc API Documentation")
    cli()

    if os.path.exists("../../../build"):
        shutil.move("html", "../../../build/epydoc")
except Exception:  # pylint: disable=broad-except
    traceback.print_exc(file=sys.stdout)
    print("Epydoc not available...not building")



#!/usr/bin/env python3
"""Pymodbus Synchronous Client Example.

An example of a single threaded synchronous client.

usage: client_sync.py [-h] [--comm {tcp,udp,serial,tls}]
                      [--framer {ascii,binary,rtu,socket,tls}]
                      [--log {critical,error,warning,info,debug}]
                      [--port PORT]
options:
  -h, --help            show this help message and exit
  --comm {tcp,udp,serial,tls}
                        "serial", "tcp", "udp" or "tls"
  --framer {ascii,binary,rtu,socket,tls}
                        "ascii", "binary", "rtu", "socket" or "tls"
  --log {critical,error,warning,info,debug}
                        "critical", "error", "warning", "info" or "debug"
  --port PORT           the port to use

The corresponding server must be started before e.g. as:
    python3 server_sync.py
"""
import argparse
import os
import logging

# --------------------------------------------------------------------------- #
# import the various client implementations
# --------------------------------------------------------------------------- #
from pymodbus.client import (
    ModbusSerialClient,
    ModbusTcpClient,
    ModbusTlsClient,
    ModbusUdpClient,
)
from pymodbus.transaction import (
    ModbusAsciiFramer,
    ModbusBinaryFramer,
    ModbusRtuFramer,
    ModbusSocketFramer,
    ModbusTlsFramer,
)


def setup_sync_client(args=None):
    """Run client setup."""
    if not args:
        args = get_commandline()
    if args.comm != "serial" and args.port:
        args.port = int(args.port)
    _logger.info("### Create client object")
    if args.comm == "tcp":
        client = ModbusTcpClient(
            "127.0.0.1",
            port=args.port,
            # Common optional paramers:
            framer=args.framer,
            #    timeout=10,
            #    retries=3,
            #    retry_on_empty=False,y
            #    close_comm_on_error=False,
            #    strict=True,
            # TCP setup parameters
            #    source_address=("localhost", 0),
        )
    elif args.comm == "udp":
        client = ModbusUdpClient(
            "localhost",
            port=args.port,
            # Common optional paramers:
            framer=args.framer,
            #    timeout=10,
            #    retries=3,
            #    retry_on_empty=False,
            #    close_comm_on_error=False,
            #    strict=True,
            # UDP setup parameters
            #    source_address=None,
        )
    elif args.comm == "serial":
        client = ModbusSerialClient(
            port=args.port,  # serial port
            # Common optional paramers:
            #    framer=ModbusRtuFramer,
            #    timeout=10,
            #    retries=3,
            #    retry_on_empty=False,
            #    close_comm_on_error=False,.
            #    strict=True,
            # Serial setup parameters
            #    baudrate=9600,
            #    bytesize=8,
            #    parity="N",
            #    stopbits=1,
            #    handle_local_echo=False,
        )
    elif args.comm == "tls":
        cwd = os.getcwd().split("/")[-1]
        if cwd == "examples":
            path = "."
        elif cwd == "test":
            path = "../examples"
        else:
            path = "examples"
        client = ModbusTlsClient(
            "localhost",
            port=args.port,
            # Common optional paramers:
            framer=args.framer,
            #    timeout=10,
            #    retries=3,
            #    retry_on_empty=False,
            #    close_comm_on_error=False,
            #    strict=True,
            # TLS setup parameters
            #    sslctx=None,
            certfile=f"{path}/certificates/pymodbus.crt",
            keyfile=f"{path}/certificates/pymodbus.key",
            #    password=None,
            server_hostname="localhost",
        )
    return client


def run_sync_client(client, modbus_calls=None):
    """Run sync client."""
    _logger.info("### Client starting")
    client.connect()
    if modbus_calls:
        modbus_calls(client)
    client.close()
    _logger.info("### End of Program")


# --------------------------------------------------------------------------- #
# Extra code, to allow commandline parameters instead of changing the code
# --------------------------------------------------------------------------- #
FORMAT = "%(asctime)-15s %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
logging.basicConfig(format=FORMAT)
_logger = logging.getLogger()


def get_commandline():
    """Read and validate command line arguments"""
    parser = argparse.ArgumentParser(
        description="Connect/disconnect a synchronous client."
    )
    parser.add_argument(
        "--comm",
        choices=["tcp", "udp", "serial", "tls"],
        help='"serial", "tcp", "udp" or "tls"',
        type=str,
    )
    parser.add_argument(
        "--framer",
        choices=["ascii", "binary", "rtu", "socket", "tls"],
        help='"ascii", "binary", "rtu", "socket" or "tls"',
        type=str,
    )
    parser.add_argument(
        "--log",
        choices=["critical", "error", "warning", "info", "debug"],
        help='"critical", "error", "warning", "info" or "debug"',
        type=str,
    )
    parser.add_argument(
        "--port",
        help="the port to use",
        type=str,
    )
    args = parser.parse_args()

    # set defaults
    comm_defaults = {
        "tcp": ["socket", 5020],
        "udp": ["socket", 5020],
        "serial": ["rtu", "/dev/ptyp0"],
        "tls": ["tls", 5020],
    }
    framers = {
        "ascii": ModbusAsciiFramer,
        "binary": ModbusBinaryFramer,
        "rtu": ModbusRtuFramer,
        "socket": ModbusSocketFramer,
        "tls": ModbusTlsFramer,
    }
    _logger.setLevel(args.log.upper() if args.log else logging.INFO)
    if not args.comm:
        args.comm = "tcp"
    if not args.framer:
        args.framer = comm_defaults[args.comm][0]
    args.port = args.port or comm_defaults[args.comm][1]
    args.framer = framers[args.framer]
    return args


if __name__ == "__main__":
    # Connect/disconnect no calls.
    testclient = setup_sync_client()
    run_sync_client(testclient)



# pylint: disable=differing-param-doc,missing-any-param-doc
"""Pymodbus Server With Updating Thread.

This is an example of having a background thread updating the
context in an SQLite4 database while the server is operating.

This scrit generates a random address range (within 0 - 65000) and a random
value and stores it in a database. It then reads the same address to verify
that the process works as expected

This can also be done with a python thread::
    from threading import Thread
    thread = Thread(target=updating_writer, args=(context,))
    thread.start()
"""
import asyncio
import logging
import random

from pymodbus.datastore import ModbusSequentialDataBlock, ModbusServerContext
from pymodbus.datastore.database import SqlSlaveContext
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.server import StartAsyncTcpServer

# --------------------------------------------------------------------------- #
# import the modbus libraries we need
# --------------------------------------------------------------------------- #
from pymodbus.version import version


# from pymodbus.transaction import ModbusRtuFramer, ModbusAsciiFramer

# --------------------------------------------------------------------------- #
# configure the service logging
# --------------------------------------------------------------------------- #
log = logging.getLogger()
log.setLevel(logging.DEBUG)

# --------------------------------------------------------------------------- #
# define your callback process
# --------------------------------------------------------------------------- #


def updating_writer(parm1):
    """Run every so often,

    and updates live values of the context which resides in an SQLite3 database.
    It should be noted that there is a race condition for the update.
    :param arguments: The input arguments to the call
    """
    log.debug("Updating the database context")
    context = parm1[0]
    readfunction = 0x03  # read holding registers
    writefunction = 0x10
    slave_id = 0x01  # slave address
    count = 50

    # import pdb; pdb.set_trace()

    rand_value = random.randint(0, 9999)  # nosec
    rand_addr = random.randint(0, 65000)  # nosec
    txt = f"Writing to datastore: {rand_addr}, {rand_value}"
    log.debug(txt)
    # import pdb; pdb.set_trace()
    context[slave_id].setValues(writefunction, rand_addr, [rand_value], update=False)
    values = context[slave_id].getValues(readfunction, rand_addr, count)
    txt = f"Values from datastore: {values}"
    log.debug(txt)


async def run_dbstore_update_server():
    """Run dbstore update server."""
    # ----------------------------------------------------------------------- #
    # initialize your data store
    # ----------------------------------------------------------------------- #

    block = ModbusSequentialDataBlock(0x00, [0] * 0xFF)
    store = SqlSlaveContext(block)

    context = ModbusServerContext(slaves={1: store}, single=False)

    # ----------------------------------------------------------------------- #
    # initialize the server information
    # ----------------------------------------------------------------------- #
    identity = ModbusDeviceIdentification(
        info_name={
            "VendorName": "pymodbus",
            "ProductCode": "PM",
            "VendorUrl": "https://github.com/riptideio/pymodbus/",
            "ProductName": "pymodbus Server",
            "ModelName": "pymodbus Server",
            "MajorMinorRevision": version.short(),
        }
    )

    # ----------------------------------------------------------------------- #
    # run the server you want
    # ----------------------------------------------------------------------- #
    time = 5  # 5 seconds delay
    loop = asyncio.get_event_loop()
    loop.start(time, now=False)  # initially delay by time
    loop.stop()
    await StartAsyncTcpServer(context, identity=identity, address=("", 5020))


if __name__ == "__main__":
    asyncio.run(run_dbstore_update_server())



#!/usr/bin/env python3
"""Pymodbus Asynchronous Processor Example.

The following is a full example of a continuous client processor. Feel
free to use it as a skeleton guide in implementing your own.
"""
# --------------------------------------------------------------------------- #
# import the necessary modules
# --------------------------------------------------------------------------- #
import logging

from twisted.internet import reactor, serialport
from twisted.internet.protocol import ClientFactory

from pymodbus.client.asynchronous.twisted import ModbusClientProtocol
from pymodbus.factory import ClientDecoder

# --------------------------------------------------------------------------- #
# Choose the framer you want to use
# --------------------------------------------------------------------------- #
# from pymodbus.transaction import ModbusBinaryFramer as ModbusFramer
# from pymodbus.transaction import ModbusAsciiFramer as ModbusFramer
from pymodbus.transaction import ModbusRtuFramer as ModbusFramer


# from pymodbus.transaction import ModbusSocketFramer as ModbusFramer

# --------------------------------------------------------------------------- #
# configure the client logging
# --------------------------------------------------------------------------- #
FORMAT = (
    "%(asctime)-15s %(threadName)-15s"
    " %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
)
logging.basicConfig(format=FORMAT)
log = logging.getLogger()
log.setLevel(logging.DEBUG)

# --------------------------------------------------------------------------- #
# state a few constants
# --------------------------------------------------------------------------- #

SERIAL_PORT = "/dev/ptyp0"
STATUS_REGS = (1, 2)
STATUS_COILS = (1, 3)
CLIENT_DELAY = 1
UNIT = 0x01


# --------------------------------------------------------------------------- #
# an example custom protocol
# --------------------------------------------------------------------------- #
# Here you can perform your main processing loop utilizing defereds and timed
# callbacks.
# --------------------------------------------------------------------------- #
class ExampleProtocol(ModbusClientProtocol):
    """Example protocol."""

    def __init__(self, framer, endpoint):
        """Initialize our custom protocol

        :param framer: The decoder to use to process messages
        :param endpoint: The endpoint to send results to
        """
        ModbusClientProtocol.__init__(self, framer)
        self.endpoint = endpoint
        log.debug("Beginning the processing loop")
        reactor.callLater(CLIENT_DELAY, self.fetch_holding_registers)

    def fetch_holding_registers(self):
        """Defer fetching holding registers"""
        log.debug("Starting the next cycle")
        data = self.read_holding_registers(*STATUS_REGS, unit=UNIT)
        data.addCallbacks(self.send_holding_registers, self.error_handler)

    def send_holding_registers(self, response):
        """Write values of holding registers, defer fetching coils

        :param response: The response to process
        """
        self.endpoint.write(response.getRegister(0))
        self.endpoint.write(response.getRegister(1))
        result = self.read_coils(*STATUS_COILS, unit=UNIT)
        result.addCallbacks(self.start_next_cycle, self.error_handler)

    def start_next_cycle(self, response):
        """Write values of coils, trigger next cycle

        :param response: The response to process
        """
        self.endpoint.write(response.getBit(0))
        self.endpoint.write(response.getBit(1))
        self.endpoint.write(response.getBit(2))
        reactor.callLater(CLIENT_DELAY, self.fetch_holding_registers)

    def error_handler(self, failure):
        """Handle any twisted errors

        :param failure: The error to handle
        """
        log.error(failure)


# --------------------------------------------------------------------------- #
# a factory for the example protocol
# --------------------------------------------------------------------------- #
# This is used to build client protocol"s if you tie into twisted"s method
# of processing. It basically produces client instances of the underlying
# protocol::
#
#     Factory(Protocol) -> ProtocolInstance
#
# It also persists data between client instances (think protocol singleton).
# --------------------------------------------------------------------------- #
class ExampleFactory(ClientFactory):
    """Example factory."""

    protocol = ExampleProtocol

    def __init__(self, framer, endpoint):
        """Remember things necessary for building a protocols"""
        self.framer = framer
        self.endpoint = endpoint

    def buildProtocol(self, _):
        """Create a protocol and start the reading cycle"""
        proto = self.protocol(self.framer, self.endpoint)
        proto.factory = self
        return proto


# --------------------------------------------------------------------------- #
# a custom client for our device
# --------------------------------------------------------------------------- #
# Twisted provides a number of helper methods for creating and starting
# clients:
# - protocol.ClientCreator
# - reactor.connectTCP
#
# How you start your client is really up to you.
# --------------------------------------------------------------------------- #
class SerialModbusClient(serialport.SerialPort):
    """Serial modbus client."""

    def __init__(self, factory, *args, **kwargs):
        """Do setup the client and start listening on the serial port

        :param factory: The factory to build clients with
        """
        protocol = factory.buildProtocol(None)
        self.decoder = ClientDecoder()
        serialport.SerialPort.__init__(self, protocol, *args, **kwargs)


# --------------------------------------------------------------------------- #
# a custom endpoint for our results
# --------------------------------------------------------------------------- #
# An example line reader, this can replace with:
# - the TCP protocol
# - a context recorder
# - a database or file recorder
# --------------------------------------------------------------------------- #
class LoggingLineReader:
    """Logging line reader."""

    def write(self, response):
        """Handle the next modbus response

        :param response: The response to process
        """
        txt = f"Read Data: {response}"
        log.info(txt)


# --------------------------------------------------------------------------- #
# start running the processor
# --------------------------------------------------------------------------- #
# This initializes the client, the framer, the factory, and starts the
# twisted event loop (the reactor). It should be noted that a number of
# things could be chanegd as one sees fit:
# - The ModbusRtuFramer could be replaced with a ModbusAsciiFramer
# - The SerialModbusClient could be replaced with reactor.connectTCP
# - The LineReader endpoint could be replaced with a database store
# --------------------------------------------------------------------------- #


def main():
    """Run Main."""
    log.debug("Initializing the client")
    framer = ModbusFramer(ClientDecoder(), client=None)
    reader = LoggingLineReader()
    factory = ExampleFactory(framer, reader)
    SerialModbusClient(factory, SERIAL_PORT, reactor)
    # factory = reactor.connectTCP("localhost", 502, factory)
    log.debug("Starting the client")
    reactor.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# pylint: disable=missing-type-doc,missing-param-doc,differing-param-doc,missing-raises-doc
"""Libmodbus Protocol Wrapper.

What follows is an example wrapper of the libmodbus library
(https://libmodbus.org/documentation/) for use with pymodbus.
There are two utilities involved here:

* LibmodbusLevel1Client

  This is simply a python wrapper around the c library. It is
  mostly a clone of the pylibmodbus implementation, but I plan
  on extending it to implement all the available protocol using
  the raw execute methods.

* LibmodbusClient

  This is just another modbus client that can be used just like
  any other client in pymodbus.

For these to work, you must have `cffi` and `libmodbus-dev` installed:

    sudo apt-get install libmodbus-dev
    pip install cffi
"""
# -------------------------------------------------------------------------- #
# import system libraries
# -------------------------------------------------------------------------- #
from cffi import FFI  # pylint: disable=import-error

from pymodbus.bit_read_message import (
    ReadCoilsResponse,
    ReadDiscreteInputsResponse,
)
from pymodbus.bit_write_message import (
    WriteMultipleCoilsResponse,
    WriteSingleCoilResponse,
)
from pymodbus.client.mixin import ModbusClientMixin
from pymodbus.constants import Defaults
from pymodbus.exceptions import ModbusException
from pymodbus.register_read_message import (
    ReadHoldingRegistersResponse,
    ReadInputRegistersResponse,
    ReadWriteMultipleRegistersResponse,
)
from pymodbus.register_write_message import (
    WriteMultipleRegistersResponse,
    WriteSingleRegisterResponse,
)


# -------------------------------------------------------------------------- #
# import pymodbus libraries
# -------------------------------------------------------------------------- #


# --------------------------------------------------------------------------- #
# create the C interface
# --------------------------------------------------------------------------- #
# * TODO add the protocol needed for the servers
# --------------------------------------------------------------------------- #

compiler = FFI()
compiler.cdef(
    """
    typedef struct _modbus modbus_t;

    int modbus_connect(modbus_t *ctx);
    int modbus_flush(modbus_t *ctx);
    void modbus_close(modbus_t *ctx);

    const char *modbus_strerror(int errnum);
    int modbus_set_slave(modbus_t *ctx, int slave);

    void modbus_get_response_timeout(modbus_t *ctx, uint32_t *to_sec, uint32_t *to_usec);
    void modbus_set_response_timeout(modbus_t *ctx, uint32_t to_sec, uint32_t to_usec);

    int modbus_read_bits(modbus_t *ctx, int addr, int nb, uint8_t *dest);
    int modbus_read_input_bits(modbus_t *ctx, int addr, int nb, uint8_t *dest);
    int modbus_read_registers(modbus_t *ctx, int addr, int nb, uint16_t *dest);
    int modbus_read_input_registers(modbus_t *ctx, int addr, int nb, uint16_t *dest);

    int modbus_write_bit(modbus_t *ctx, int coil_addr, int status);
    int modbus_write_bits(modbus_t *ctx, int addr, int nb, const uint8_t *data);
    int modbus_write_register(modbus_t *ctx, int reg_addr, int value);
    int modbus_write_registers(modbus_t *ctx, int addr, int nb, const uint16_t *data);
    int modbus_write_and_read_registers(modbus_t *ctx, int write_addr, int write_nb,
                                        const uint16_t *src, int read_addr, int read_nb, uint16_t *dest);

    int modbus_mask_write_register(modbus_t *ctx, int addr, uint16_t and_mask, uint16_t or_mask);
    int modbus_send_raw_request(modbus_t *ctx, uint8_t *raw_req, int raw_req_length);

    float modbus_get_float(const uint16_t *src);
    void modbus_set_float(float f, uint16_t *dest);

    modbus_t* modbus_new_tcp(const char *ip_address, int port);
    modbus_t* modbus_new_rtu(const char *device, int baud, char parity, int data_bit, int stop_bit);
    void modbus_free(modbus_t *ctx);

    int modbus_receive(modbus_t *ctx, uint8_t *req);
    int modbus_receive_from(modbus_t *ctx, int sockfd, uint8_t *req);
    int modbus_receive_confirmation(modbus_t *ctx, uint8_t *rsp);
"""
)
LIB = compiler.dlopen("modbus")  # create our bindings

# -------------------------------------------------------------------------- #
# helper utilities
# -------------------------------------------------------------------------- #


def get_float(data):
    """Get float."""
    return LIB.modbus_get_float(data)


def set_float(value, data):
    """Set float."""
    LIB.modbus_set_float(value, data)


def cast_to_int16(data):
    """Cast to int16."""
    return int(compiler.cast("int16_t", data))


def cast_to_int32(data):
    """Cast to int32."""
    return int(compiler.cast("int32_t", data))


class NotImplementedException(Exception):
    """Not implemented exception."""


# -------------------------------------------------------------------------- #
# level1 client
# -------------------------------------------------------------------------- #


class LibmodbusLevel1Client:
    """A raw wrapper around the libmodbus c library.

    Feel free to use it if you want increased performance and don"t mind the
    entire protocol not being implemented.
    """

    @classmethod
    def create_tcp_client(cls, my_host="127.0.0.1", my_port=Defaults.TcpPort):
        """Create a TCP modbus client for the supplied parameters.

        :param host: The host to connect to
        :param port: The port to connect to on that host
        :returns: A new level1 client
        """
        my_client = LIB.modbus_new_tcp(my_host.encode(), my_port)
        return cls(my_client)

    @classmethod
    def create_rtu_client(cls, **kwargs):
        """Create a TCP modbus client for the supplied parameters.

        :param port: The serial port to attach to
        :param stopbits: The number of stop bits to use
        :param bytesize: The bytesize of the serial messages
        :param parity: Which kind of parity to use
        :param baudrate: The baud rate to use for the serial device
        :returns: A new level1 client
        """
        my_port = kwargs.get("port", "/dev/ttyS0")
        baudrate = kwargs.get("baud", Defaults.Baudrate)
        parity = kwargs.get("parity", Defaults.Parity)
        bytesize = kwargs.get("bytesize", Defaults.Bytesize)
        stopbits = kwargs.get("stopbits", Defaults.Stopbits)
        my_client = LIB.modbus_new_rtu(my_port, baudrate, parity, bytesize, stopbits)
        return cls(my_client)

    def __init__(self, my_client):
        """Initialize a new instance of the LibmodbusLevel1Client.

        This method should not be used, instead new instances should be created
        using the two supplied factory methods:

        * LibmodbusLevel1Client.create_rtu_client(...)
        * LibmodbusLevel1Client.create_tcp_client(...)

        :param client: The underlying client instance to operate with.
        """
        self.client = my_client
        self.slave = Defaults.Slave

    def set_slave(self, slave):
        """Set the current slave to operate against.

        :param slave: The new slave to operate against
        :returns: The resulting slave to operate against
        """
        self.slave = self._execute(  # pylint: disable=no-member
            LIB.modbus_set_slave, slave
        )
        return self.slave

    def connect(self):
        """Attempt to connect to the client target.

        :returns: True if successful, throws otherwise
        """
        return not self.__execute(LIB.modbus_connect)

    def flush(self):
        """Discard the existing bytes on the wire.

        :returns: The number of flushed bytes, or throws
        """
        return self.__execute(LIB.modbus_flush)

    def close(self):
        """Close and frees the underlying connection and context structure.

        :returns: Always True
        """
        LIB.modbus_close(self.client)
        LIB.modbus_free(self.client)
        return True

    def __execute(self, command, *args):
        """Run the supplied command against the currently instantiated client with the supplied arguments.

        This will make sure to correctly handle resulting errors.

        :param command: The command to execute against the context
        :param *args: The arguments for the given command
        :returns: The result of the operation unless -1 which throws
        """
        if (result := command(self.client, *args)) == -1:
            message = LIB.modbus_strerror(compiler.errno)
            raise ModbusException(compiler.string(message))
        return result

    def read_bits(self, address, count=1):
        """Read bits.

        :param address: The starting address to read from
        :param count: The number of coils to read
        :returns: The resulting bits
        """
        result = compiler.new("uint8_t[]", count)
        self.__execute(LIB.modbus_read_bits, address, count, result)
        return result

    def read_input_bits(self, address, count=1):
        """Read input bits.

        :param address: The starting address to read from
        :param count: The number of discretes to read
        :returns: The resulting bits
        """
        result = compiler.new("uint8_t[]", count)
        self.__execute(LIB.modbus_read_input_bits, address, count, result)
        return result

    def write_bit(self, address, value):
        """Write bit.

        :param address: The starting address to write to
        :param value: The value to write to the specified address
        :returns: The number of written bits
        """
        return self.__execute(LIB.modbus_write_bit, address, value)

    def write_bits(self, address, values):
        """Write bits.

        :param address: The starting address to write to
        :param values: The values to write to the specified address
        :returns: The number of written bits
        """
        count = len(values)
        return self.__execute(LIB.modbus_write_bits, address, count, values)

    def write_register(self, address, value):
        """Write register.

        :param address: The starting address to write to
        :param value: The value to write to the specified address
        :returns: The number of written registers
        """
        return self.__execute(LIB.modbus_write_register, address, value)

    def write_registers(self, address, values):
        """Write registers.

        :param address: The starting address to write to
        :param values: The values to write to the specified address
        :returns: The number of written registers
        """
        count = len(values)
        return self.__execute(LIB.modbus_write_registers, address, count, values)

    def read_registers(self, address, count=1):
        """Read registers.

        :param address: The starting address to read from
        :param count: The number of registers to read
        :returns: The resulting read registers
        """
        result = compiler.new("uint16_t[]", count)
        self.__execute(LIB.modbus_read_registers, address, count, result)
        return result

    def read_input_registers(self, address, count=1):
        """Read input registers.

        :param address: The starting address to read from
        :param count: The number of registers to read
        :returns: The resulting read registers
        """
        result = compiler.new("uint16_t[]", count)
        self.__execute(LIB.modbus_read_input_registers, address, count, result)
        return result

    def read_and_write_registers(
        self, read_address, read_count, write_address, write_registers
    ):
        """Read/write registers.

        :param read_address: The address to start reading from
        :param read_count: The number of registers to read from address
        :param write_address: The address to start writing to
        :param write_registers: The registers to write to the specified address
        :returns: The resulting read registers
        """
        write_count = len(write_registers)
        read_result = compiler.new("uint16_t[]", read_count)
        self.__execute(
            LIB.modbus_write_and_read_registers,
            write_address,
            write_count,
            write_registers,
            read_address,
            read_count,
            read_result,
        )
        return read_result


# -------------------------------------------------------------------------- #
# level2 client
# -------------------------------------------------------------------------- #


class LibmodbusClient(ModbusClientMixin):
    """A facade around the raw level 1 libmodbus client.

    that implements the pymodbus protocol on top of the lower level client.
    """

    # ----------------------------------------------------------------------- #
    # these are used to convert from the pymodbus request types to the
    # libmodbus operations (overloaded operator).
    # ----------------------------------------------------------------------- #

    __methods = {
        "ReadCoilsRequest": lambda c, r: c.read_bits(r.address, r.count),
        "ReadDiscreteInputsRequest": lambda c, r: c.read_input_bits(r.address, r.count),
        "WriteSingleCoilRequest": lambda c, r: c.write_bit(r.address, r.value),
        "WriteMultipleCoilsRequest": lambda c, r: c.write_bits(r.address, r.values),
        "WriteSingleRegisterRequest": lambda c, r: c.write_register(r.address, r.value),
        "WriteMultipleRegistersRequest": lambda c, r: c.write_registers(
            r.address, r.values
        ),
        "ReadHoldingRegistersRequest": lambda c, r: c.read_registers(
            r.address, r.count
        ),
        "ReadInputRegistersRequest": lambda c, r: c.read_input_registers(
            r.address, r.count
        ),
        "ReadWriteMultipleRegistersRequest": lambda c, r: c.read_and_write_registers(
            r.read_address, r.read_count, r.write_address, r.write_registers
        ),
    }

    # ----------------------------------------------------------------------- #
    # these are used to convert from the libmodbus result to the
    # pymodbus response type
    # ----------------------------------------------------------------------- #

    __adapters = {
        "ReadCoilsRequest": lambda tx, rx: ReadCoilsResponse(list(rx)),
        "ReadDiscreteInputsRequest": lambda tx, rx: ReadDiscreteInputsResponse(
            list(rx)
        ),
        "WriteSingleCoilRequest": lambda tx, rx: WriteSingleCoilResponse(
            tx.address, rx
        ),
        "WriteMultipleCoilsRequest": lambda tx, rx: WriteMultipleCoilsResponse(
            tx.address, rx
        ),
        "WriteSingleRegisterRequest": lambda tx, rx: WriteSingleRegisterResponse(
            tx.address, rx
        ),
        "WriteMultipleRegistersRequest": lambda tx, rx: WriteMultipleRegistersResponse(
            tx.address, rx
        ),
        "ReadHoldingRegistersRequest": lambda tx, rx: ReadHoldingRegistersResponse(
            list(rx)
        ),
        "ReadInputRegistersRequest": lambda tx, rx: ReadInputRegistersResponse(
            list(rx)
        ),
        "ReadWriteMultipleRegistersRequest": lambda tx, rx: ReadWriteMultipleRegistersResponse(
            list(rx)
        ),
    }

    def __init__(self, my_client):
        """Initialize a new instance of the LibmodbusClient.

        This should be initialized with one of the LibmodbusLevel1Client instances:

        * LibmodbusLevel1Client.create_rtu_client(...)
        * LibmodbusLevel1Client.create_tcp_client(...)
        :param client: The underlying client instance to operate with.
        """
        self.client = my_client

    # ----------------------------------------------------------------------- #
    # We use the client mixin to implement the api methods which are all
    # forwarded to this method. It is implemented using the previously
    # defined lookup tables. Any method not defined simply throws.
    # ----------------------------------------------------------------------- #

    def execute(self, request):
        """Execute the supplied request against the server.

        :param request: The request to process
        :returns: The result of the request execution
        """
        if self.client.slave != request.unit_id:
            self.client.set_slave(request.unit_id)

        method = request.__class__.__name__
        operation = self.__methods.get(method, None)
        adapter = self.__adapters.get(method, None)

        if not operation or not adapter:
            raise NotImplementedException("Method not implemented: " + operation)

        response = operation(self.client, request)
        return adapter(request, response)

    # ----------------------------------------------------------------------- #
    # Other methods can simply be forwarded using the decorator pattern
    # ----------------------------------------------------------------------- #

    def connect(self):
        """Connect."""
        return self.client.connect()

    def close(self):
        """Close."""
        return self.client.close()

    # ----------------------------------------------------------------------- #
    # magic methods
    # ----------------------------------------------------------------------- #

    def __enter__(self):
        """Implement the client with enter block

        :returns: The current instance of the client
        """
        self.client.connect()
        return self

    def __exit__(self, klass, value, traceback):
        """Implement the client with exit block"""
        self.client.close()


# -------------------------------------------------------------------------- #
# main example runner
# -------------------------------------------------------------------------- #


if __name__ == "__main__":

    # create our low level client
    host = "127.0.0.1"  # pylint: disable=invalid-name
    port = 502  # pylint: disable=invalid-name
    protocol = LibmodbusLevel1Client.create_tcp_client(host, port)

    # operate with our high level client
    with LibmodbusClient(protocol) as client:
        registers = client.write_registers(0, [13, 12, 11])
        print(registers)
        registers = client.read_holding_registers(0, 10)
        print(registers.registers)



"""Test version."""
import unittest

from pymodbus.version import Version, version


class ModbusVersionTest(unittest.TestCase):
    """Unittest for the pymodbus._version code."""

    def setUp(self):
        """Initialize the test environment"""

    def tearDown(self):
        """Clean up the test environment"""

    def test_version_class(self):
        """Test version class."""
        version = Version("test", 1, 2, 3, "sometag")
        self.assertEqual(version.short(), "1.2.3.sometag")
        self.assertEqual(str(version), "[test, version 1.2.3.sometag]")

    def test_current_version(self):
        """Test current version"""
        self.assertEqual(version.package, "pymodbus")



"""Modbus Remote Events.

An event byte returned by the Get Communications Event Log function
can be any one of four types. The type is defined by bit 7
(the high-order bit) in each byte. It may be further defined by bit 6.
"""
# pylint: disable=missing-type-doc
from pymodbus.exceptions import NotImplementedException, ParameterException
from pymodbus.utilities import pack_bitstring, unpack_bitstring


class ModbusEvent:
    """Define modbus events."""

    def encode(self):
        """Encode the status bits to an event message.

        :raises NotImplementedException:
        """
        raise NotImplementedException()

    def decode(self, event):
        """Decode the event message to its status bits.

        :param event: The event to decode
        :raises NotImplementedException:
        """
        raise NotImplementedException()


class RemoteReceiveEvent(ModbusEvent):
    """Remote device MODBUS Receive Event.

    The remote device stores this type of event byte when a query message
    is received. It is stored before the remote device processes the message.
    This event is defined by bit 7 set to logic "1". The other bits will be
    set to a logic "1" if the corresponding condition is TRUE. The bit layout
    is::

        Bit Contents
        ----------------------------------
        0   Not Used
        2   Not Used
        3   Not Used
        4   Character Overrun
        5   Currently in Listen Only Mode
        6   Broadcast Receive
        7   1
    """

    def __init__(self, **kwargs):
        """Initialize a new event instance."""
        self.overrun = kwargs.get("overrun", False)
        self.listen = kwargs.get("listen", False)
        self.broadcast = kwargs.get("broadcast", False)

    def encode(self):
        """Encode the status bits to an event message.

        :returns: The encoded event message
        """
        bits = [False] * 3
        bits += [self.overrun, self.listen, self.broadcast, True]
        packet = pack_bitstring(bits)
        return packet

    def decode(self, event):
        """Decode the event message to its status bits.

        :param event: The event to decode
        """
        bits = unpack_bitstring(event)
        self.overrun = bits[4]
        self.listen = bits[5]
        self.broadcast = bits[6]


class RemoteSendEvent(ModbusEvent):
    """Remote device MODBUS Send Event.

    The remote device stores this type of event byte when it finishes
    processing a request message. It is stored if the remote device
    returned a normal or exception response, or no response.

    This event is defined by bit 7 set to a logic "0", with bit 6 set to a "1".
    The other bits will be set to a logic "1" if the corresponding
    condition is TRUE. The bit layout is::

        Bit Contents
        -----------------------------------------------------------
        0   Read Exception Sent (Exception Codes 1-3)
        1   Slave Abort Exception Sent (Exception Code 4)
        2   Slave Busy Exception Sent (Exception Codes 5-6)
        3   Slave Program NAK Exception Sent (Exception Code 7)
        4   Write Timeout Error Occurred
        5   Currently in Listen Only Mode
        6   1
        7   0
    """

    def __init__(self, **kwargs):
        """Initialize a new event instance."""
        self.read = kwargs.get("read", False)
        self.slave_abort = kwargs.get("slave_abort", False)
        self.slave_busy = kwargs.get("slave_busy", False)
        self.slave_nak = kwargs.get("slave_nak", False)
        self.write_timeout = kwargs.get("write_timeout", False)
        self.listen = kwargs.get("listen", False)

    def encode(self):
        """Encode the status bits to an event message.

        :returns: The encoded event message
        """
        bits = [
            self.read,
            self.slave_abort,
            self.slave_busy,
            self.slave_nak,
            self.write_timeout,
            self.listen,
        ]
        bits += [True, False]
        packet = pack_bitstring(bits)
        return packet

    def decode(self, event):
        """Decode the event message to its status bits.

        :param event: The event to decode
        """
        # todo fix the start byte count # pylint: disable=fixme
        bits = unpack_bitstring(event)
        self.read = bits[0]
        self.slave_abort = bits[1]
        self.slave_busy = bits[2]
        self.slave_nak = bits[3]
        self.write_timeout = bits[4]
        self.listen = bits[5]


class EnteredListenModeEvent(ModbusEvent):
    """Enter Remote device Listen Only Mode

    The remote device stores this type of event byte when it enters
    the Listen Only Mode. The event is defined by a content of 04 hex.
    """

    value = 0x04
    __encoded = b"\x04"

    def encode(self):
        """Encode the status bits to an event message.

        :returns: The encoded event message
        """
        return self.__encoded

    def decode(self, event):
        """Decode the event message to its status bits.

        :param event: The event to decode
        :raises ParameterException:
        """
        if event != self.__encoded:
            raise ParameterException("Invalid decoded value")


class CommunicationRestartEvent(ModbusEvent):
    """Restart remote device Initiated Communication.

    The remote device stores this type of event byte when its communications
    port is restarted. The remote device can be restarted by the Diagnostics
    function (code 08), with sub-function Restart Communications Option
    (code 00 01).

    That function also places the remote device into a "Continue on Error"
    or "Stop on Error" mode. If the remote device is placed  into "Continue on
    Error" mode, the event byte is added to the existing event log. If the
    remote device is placed into "Stop on Error" mode, the byte is added to
    the log and the rest of the log is cleared to zeros.

    The event is defined by a content of zero.
    """

    value = 0x00
    __encoded = b"\x00"

    def encode(self):
        """Encode the status bits to an event message.

        :returns: The encoded event message
        """
        return self.__encoded

    def decode(self, event):
        """Decode the event message to its status bits.

        :param event: The event to decode
        :raises ParameterException:
        """
        if event != self.__encoded:
            raise ParameterException("Invalid decoded value")



#!/usr/bin/env python3
"""Pymodbus Synchronous Client Example to showcase Device Information.

This client demonstrates the use of Device Information to get information
about servers connected to the client. This is part of the MODBUS specification,
and uses the MEI 0x2B 0x0E request / response.
"""
import logging

# --------------------------------------------------------------------------- #
# import the various server implementations
# --------------------------------------------------------------------------- #
from pymodbus.client import ModbusTcpClient as ModbusClient
from pymodbus.device import ModbusDeviceIdentification

# --------------------------------------------------------------------------- #
# import the request
# --------------------------------------------------------------------------- #
from pymodbus.mei_message import ReadDeviceInformationRequest


# from pymodbus.client import ModbusUdpClient as ModbusClient
# from pymodbus.client import ModbusSerialClient as ModbusClient


# --------------------------------------------------------------------------- #
# configure the client logging
# --------------------------------------------------------------------------- #
FORMAT = (
    "%(asctime)-15s %(threadName)-15s "
    "%(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
)
logging.basicConfig(format=FORMAT)
log = logging.getLogger()
log.setLevel(logging.DEBUG)

UNIT = 0x1


def run_sync_client():
    """Run sync client."""
    # ------------------------------------------------------------------------#
    # choose the client you want
    # ------------------------------------------------------------------------#
    # make sure to start an implementation to hit against. For this
    # you can use an existing device, the reference implementation in the tools
    # directory, or start a pymodbus server.
    #
    # If you use the UDP or TCP clients, you can override the framer being used
    # to use a custom implementation (say RTU over TCP). By default they use
    # the socket framer::
    #
    #    client = ModbusClient("localhost", port=5020, framer=ModbusRtuFramer)
    #
    # It should be noted that you can supply an ipv4 or an ipv6 host address
    # for both the UDP and TCP clients.
    #
    # There are also other options that can be set on the client that controls
    # how transactions are performed. The current ones are:
    #
    # * retries - Specify how many retries to allow per transaction (default=3)
    # * retry_on_empty - Is an empty response a retry (default = False)
    # * source_address - Specifies the TCP source address to bind to
    #
    # Here is an example of using these options::
    #
    #    client = ModbusClient("localhost", retries=3, retry_on_empty=True)
    # ------------------------------------------------------------------------#
    client = ModbusClient("localhost", port=5020)
    # from pymodbus.transaction import ModbusRtuFramer
    # client = ModbusClient("localhost", port=5020, framer=ModbusRtuFramer)
    # client = ModbusClient(method="binary", port="/dev/ptyp0", timeout=1)
    # client = ModbusClient(method="ascii", port="/dev/ptyp0", timeout=1)
    # client = ModbusClient(method="rtu", port="/dev/ptyp0", timeout=1,
    #                       baudrate=9600)
    client.connect()

    # ------------------------------------------------------------------------#
    # specify slave to query
    # ------------------------------------------------------------------------#
    # The slave to query is specified in an optional parameter for each
    # individual request. This can be done by specifying the `unit` parameter
    # which defaults to `0x00`
    # ----------------------------------------------------------------------- #
    log.debug("Reading Device Information")
    information = {}
    rr = None

    while not rr or rr.more_follows:
        next_object_id = rr.next_object_id if rr else 0
        rq = ReadDeviceInformationRequest(
            read_code=0x03, unit=UNIT, object_id=next_object_id
        )
        rr = client.execute(rq)
        information.update(rr.information)
        log.debug(rr)

    print("Device Information : ")
    for (
        key
    ) in (
        information.keys()
    ):  # pylint: disable=consider-iterating-dictionary,consider-using-dict-items
        print(key, information[key])

    # ----------------------------------------------------------------------- #
    # You can also have the information parsed through the
    # ModbusDeviceIdentificiation class, which gets you a more usable way
    # to access the Basic and Regular device information objects which are
    # specifically listed in the Modbus specification
    # ----------------------------------------------------------------------- #
    device_id = ModbusDeviceIdentification(info=information)
    print("Product Name : ", device_id.ProductName)

    # ----------------------------------------------------------------------- #
    # close the client
    # ----------------------------------------------------------------------- #
    client.close()


if __name__ == "__main__":
    run_sync_client()



"""Test framers."""
from unittest.mock import Mock, patch

import pytest

from pymodbus.bit_read_message import ReadCoilsRequest
from pymodbus.client import ModbusBaseClient
from pymodbus.exceptions import ModbusIOException
from pymodbus.factory import ClientDecoder
from pymodbus.framer.ascii_framer import ModbusAsciiFramer
from pymodbus.framer.binary_framer import ModbusBinaryFramer
from pymodbus.framer.rtu_framer import ModbusRtuFramer
from pymodbus.utilities import ModbusTransactionState


TEST_MESSAGE = b"\x00\x01\x00\x01\x00\n\xec\x1c"


@pytest.fixture
def rtu_framer():
    """RTU framer."""
    return ModbusRtuFramer(ClientDecoder())


@pytest.fixture
def ascii_framer():
    """Ascii framer."""
    return ModbusAsciiFramer(ClientDecoder())


# @pytest.fixture
# def binary_framer():
#     """Binary framer."""
#     return ModbusBinaryFramer(ClientDecoder())


@pytest.mark.parametrize(
    "framer",
    [
        ModbusRtuFramer,
        ModbusAsciiFramer,
        ModbusBinaryFramer,
    ],
)
def test_framer_initialization(framer):
    """Test framer initialization."""
    decoder = ClientDecoder()
    framer = framer(decoder)
    assert framer.client is None  # nosec
    assert framer._buffer == b""  # nosec pylint: disable=protected-access
    assert framer.decoder == decoder  # nosec
    if isinstance(framer, ModbusAsciiFramer):
        assert framer._header == {  # nosec pylint: disable=protected-access
            "lrc": "0000",
            "len": 0,
            "uid": 0x00,
        }
        assert framer._hsize == 0x02  # nosec pylint: disable=protected-access
        assert framer._start == b":"  # nosec pylint: disable=protected-access
        assert framer._end == b"\r\n"  # nosec pylint: disable=protected-access
    elif isinstance(framer, ModbusRtuFramer):
        assert framer._header == {  # nosec pylint: disable=protected-access
            "uid": 0x00,
            "len": 0,
            "crc": b"\x00\x00",
        }
        assert framer._hsize == 0x01  # nosec pylint: disable=protected-access
        assert framer._end == b"\x0d\x0a"  # nosec pylint: disable=protected-access
        assert framer._min_frame_size == 4  # nosec pylint: disable=protected-access
    else:
        assert framer._header == {  # nosec pylint: disable=protected-access
            "crc": 0x0000,
            "len": 0,
            "uid": 0x00,
        }
        assert framer._hsize == 0x01  # nosec pylint: disable=protected-access
        assert framer._start == b"\x7b"  # nosec pylint: disable=protected-access
        assert framer._end == b"\x7d"  # nosec pylint: disable=protected-access
        assert framer._repeat == [  # nosec pylint: disable=protected-access
            b"}"[0],
            b"{"[0],
        ]


@pytest.mark.parametrize("data", [(b"", {}), (b"abcd", {"fcode": 98, "unit": 97})])
def test_decode_data(rtu_framer, data):  # pylint: disable=redefined-outer-name
    """Test decode data."""
    data, expected = data
    decoded = rtu_framer.decode_data(data)
    assert decoded == expected  # nosec


@pytest.mark.parametrize(
    "data",
    [
        (b"", False),
        (b"\x02\x01\x01\x00Q\xcc", True),
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD", True),  # valid frame
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAC", False),  # invalid frame CRC
    ],
)
def test_check_frame(rtu_framer, data):  # pylint: disable=redefined-outer-name
    """Test check frame."""
    data, expected = data
    rtu_framer._buffer = data  # pylint: disable=protected-access
    assert expected == rtu_framer.checkFrame()  # nosec


@pytest.mark.parametrize(
    "data",
    [
        (b"", {"uid": 0x00, "len": 0, "crc": b"\x00\x00"}, b""),
        (b"abcd", {"uid": 0x00, "len": 2, "crc": b"\x00\x00"}, b"cd"),
        (
            b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD\x12\x03",  # real case, frame size is 11
            {"uid": 0x00, "len": 11, "crc": b"\x00\x00"},
            b"\x12\x03",
        ),
    ],
)
def test_rtu_advance_framer(rtu_framer, data):  # pylint: disable=redefined-outer-name
    """Test rtu advance framer."""
    before_buf, before_header, after_buf = data

    rtu_framer._buffer = before_buf  # pylint: disable=protected-access
    rtu_framer._header = before_header  # pylint: disable=protected-access
    rtu_framer.advanceFrame()
    assert rtu_framer._header == {  # nosec pylint: disable=protected-access
        "uid": 0x00,
        "len": 0,
        "crc": b"\x00\x00",
    }
    assert rtu_framer._buffer == after_buf  # nosec pylint: disable=protected-access


@pytest.mark.parametrize("data", [b"", b"abcd"])
def test_rtu_reset_framer(rtu_framer, data):  # pylint: disable=redefined-outer-name
    """Test rtu reset framer."""
    rtu_framer._buffer = data  # pylint: disable=protected-access
    rtu_framer.resetFrame()
    assert rtu_framer._header == {  # nosec pylint: disable=protected-access
        "uid": 0x00,
        "len": 0,
        "crc": b"\x00\x00",
    }
    assert rtu_framer._buffer == b""  # nosec pylint: disable=protected-access


@pytest.mark.parametrize(
    "data",
    [
        (b"", False),
        (b"\x11", False),
        (b"\x11\x03", False),
        (b"\x11\x03\x06", False),
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49", False),
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD", True),
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD\xAB\xCD", True),
    ],
)
def test_is_frame_ready(rtu_framer, data):  # pylint: disable=redefined-outer-name
    """Test is frame ready."""
    data, expected = data
    rtu_framer._buffer = data  # pylint: disable=protected-access
    # rtu_framer.advanceFrame()
    assert rtu_framer.isFrameReady() == expected  # nosec


@pytest.mark.parametrize(
    "data",
    [
        b"",
        b"\x11",
        b"\x11\x03",
        b"\x11\x03\x06",
        b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x43",
    ],
)
def test_rtu_populate_header_fail(
    rtu_framer, data
):  # pylint: disable=redefined-outer-name
    """Test rtu populate header fail."""
    with pytest.raises(IndexError):
        rtu_framer.populateHeader(data)


@pytest.mark.parametrize(
    "data",
    [
        (
            b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD",
            {"crc": b"\x49\xAD", "uid": 17, "len": 11},
        ),
        (
            b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD\x11\x03",
            {"crc": b"\x49\xAD", "uid": 17, "len": 11},
        ),
    ],
)
def test_rtu_populate_header(rtu_framer, data):  # pylint: disable=redefined-outer-name
    """Test rtu populate header."""
    buffer, expected = data
    rtu_framer.populateHeader(buffer)
    assert rtu_framer._header == expected  # nosec pylint: disable=protected-access


def test_add_to_frame(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test add to frame."""
    assert rtu_framer._buffer == b""  # nosec pylint: disable=protected-access
    rtu_framer.addToFrame(b"abcd")
    assert rtu_framer._buffer == b"abcd"  # nosec pylint: disable=protected-access


def test_get_frame(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test get frame."""
    rtu_framer.addToFrame(b"\x02\x01\x01\x00Q\xcc")
    rtu_framer.populateHeader(b"\x02\x01\x01\x00Q\xcc")
    assert rtu_framer.getFrame() == b"\x01\x01\x00"  # nosec


def test_populate_result(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test populate result."""
    rtu_framer._header["uid"] = 255  # pylint: disable=protected-access
    result = Mock()
    rtu_framer.populateResult(result)
    assert result.unit_id == 255  # nosec


@pytest.mark.parametrize(
    "data",
    [
        (b"\x11", 17, False, False),  # not complete frame
        (b"\x11\x03", 17, False, False),  # not complete frame
        (b"\x11\x03\x06", 17, False, False),  # not complete frame
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43", 17, False, False),  # not complete frame
        (
            b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40",
            17,
            False,
            False,
        ),  # not complete frame
        (
            b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49",
            17,
            False,
            False,
        ),  # not complete frame
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAC", 17, True, False),  # bad crc
        (
            b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD",
            17,
            False,
            True,
        ),  # good frame
        (
            b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD",
            16,
            True,
            False,
        ),  # incorrect unit id
        (b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD\x11\x03", 17, False, True),
        # good frame + part of next frame
    ],
)
def test_rtu_incoming_packet(rtu_framer, data):  # pylint: disable=redefined-outer-name
    """Test rtu process incoming packet."""
    buffer, units, reset_called, process_called = data

    with patch.object(rtu_framer, "_process", wraps=rtu_framer._process) as mock_process, patch.object(  # pylint: disable=protected-access
        rtu_framer, "resetFrame", wraps=rtu_framer.resetFrame
    ) as mock_reset:
        rtu_framer.processIncomingPacket(buffer, Mock(), units)
        assert mock_process.call_count == (1 if process_called else 0)  # nosec
        assert mock_reset.call_count == (1 if reset_called else 0)  # nosec


def test_build_packet(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test build packet."""
    message = ReadCoilsRequest(1, 10)
    assert rtu_framer.buildPacket(message) == TEST_MESSAGE  # nosec


def test_send_packet(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test send packet."""
    message = TEST_MESSAGE
    client = ModbusBaseClient(framer=ModbusRtuFramer)
    client.state = ModbusTransactionState.TRANSACTION_COMPLETE
    client.silent_interval = 1
    client.last_frame_end = 1
    client.params.timeout = 0.25
    client.idle_time = Mock(return_value=1)
    client.send = Mock(return_value=len(message))
    rtu_framer.client = client
    assert rtu_framer.sendPacket(message) == len(message)  # nosec
    client.state = ModbusTransactionState.PROCESSING_REPLY
    assert rtu_framer.sendPacket(message) == len(message)  # nosec


def test_recv_packet(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test receive packet."""
    message = TEST_MESSAGE
    client = Mock()
    client.recv.return_value = message
    rtu_framer.client = client
    assert rtu_framer.recvPacket(len(message)) == message  # nosec


def test_process(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test process."""

    rtu_framer._buffer = TEST_MESSAGE  # pylint: disable=protected-access
    with pytest.raises(ModbusIOException):
        rtu_framer._process(None)  # pylint: disable=protected-access


def test_get_raw_frame(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test get raw frame."""
    rtu_framer._buffer = TEST_MESSAGE  # pylint: disable=protected-access
    assert (
        rtu_framer.getRawFrame()
        == rtu_framer._buffer  # nosec pylint: disable=protected-access
    )


def test_validate_unit_id(rtu_framer):  # pylint: disable=redefined-outer-name
    """Test validate unit."""
    rtu_framer.populateHeader(TEST_MESSAGE)
    assert rtu_framer._validate_unit_id(  # nosec pylint: disable=protected-access
        [0], False
    )
    assert rtu_framer._validate_unit_id(  # nosec pylint: disable=protected-access
        [1], True
    )


@pytest.mark.parametrize("data", [b":010100010001FC\r\n", b""])
def test_decode_ascii_data(ascii_framer, data):  # pylint: disable=redefined-outer-name
    """Test decode ascii."""
    data = ascii_framer.decode_data(data)
    assert isinstance(data, dict)  # nosec
    if data:
        assert data.get("unit") == 1  # nosec
        assert data.get("fcode") == 1  # nosec
    else:
        assert not data  # nosec



"""Datastore using redis."""
# pylint: disable=missing-type-doc
import logging

import redis

from pymodbus.interfaces import IModbusSlaveContext
from pymodbus.utilities import pack_bitstring, unpack_bitstring


# ---------------------------------------------------------------------------#
#  Logging
# ---------------------------------------------------------------------------#
_logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------#
#  Context
# ---------------------------------------------------------------------------#
class RedisSlaveContext(IModbusSlaveContext):
    """This is a modbus slave context using redis as a backing store."""

    def __init__(self, **kwargs):
        """Initialize the datastores.

        :param host: The host to connect to
        :param port: The port to connect to
        :param prefix: A prefix for the keys
        """
        host = kwargs.get("host", "localhost")
        port = kwargs.get("port", 6379)
        self.prefix = kwargs.get("prefix", "pymodbus")
        self.client = kwargs.get("client", redis.Redis(host=host, port=port))
        self._build_mapping()

    def __str__(self):
        """Return a string representation of the context.

        :returns: A string representation of the context
        """
        return f"Redis Slave Context {self.client}"

    def reset(self):
        """Reset all the datastores to their default values."""
        self.client.flushall()

    def validate(self, fx, address, count=1):
        """Validate the request to make sure it is in range.

        :param fx: The function we are working with
        :param address: The starting address
        :param count: The number of values to test
        :returns: True if the request in within range, False otherwise
        """
        address = address + 1  # section 4.4 of specification
        txt = f"validate[{fx}] {address}:{count}"
        _logger.debug(txt)
        return self._val_callbacks[self.decode(fx)](address, count)

    def getValues(self, fx, address, count=1):
        """Get `count` values from datastore.

        :param fx: The function we are working with
        :param address: The starting address
        :param count: The number of values to retrieve
        :returns: The requested values from a:a+c
        """
        address = address + 1  # section 4.4 of specification
        txt = f"getValues[{fx}] {address}:{count}"
        _logger.debug(txt)
        return self._get_callbacks[self.decode(fx)](address, count)

    def setValues(self, fx, address, values):
        """Set the datastore with the supplied values.

        :param fx: The function we are working with
        :param address: The starting address
        :param values: The new values to be set
        """
        address = address + 1  # section 4.4 of specification
        txt = f"setValues[{fx}] {address}:{len(values)}"
        _logger.debug(txt)
        self._set_callbacks[self.decode(fx)](address, values)

    # --------------------------------------------------------------------------#
    #  Redis Helper Methods
    # --------------------------------------------------------------------------#
    def _get_prefix(self, key):
        """Abstract getting bit values.

        :param key: The key prefix to use
        :returns: The key prefix to redis
        """
        return f"{self.prefix}:{key}"

    def _build_mapping(self):
        """Build the function code mapper."""
        self._val_callbacks = {
            "d": lambda o, c: self._val_bit("d", o, c),
            "c": lambda o, c: self._val_bit("c", o, c),
            "h": lambda o, c: self._val_reg("h", o, c),
            "i": lambda o, c: self._val_reg("i", o, c),
        }
        self._get_callbacks = {
            "d": lambda o, c: self._get_bit("d", o, c),
            "c": lambda o, c: self._get_bit("c", o, c),
            "h": lambda o, c: self._get_reg("h", o, c),
            "i": lambda o, c: self._get_reg("i", o, c),
        }
        self._set_callbacks = {
            "d": lambda o, v: self._set_bit("d", o, v),
            "c": lambda o, v: self._set_bit("c", o, v),
            "h": lambda o, v: self._set_reg("h", o, v),
            "i": lambda o, v: self._set_reg("i", o, v),
        }

    # --------------------------------------------------------------------------#
    #  Redis discrete implementation
    # --------------------------------------------------------------------------#
    _bit_size = 16
    _bit_default = "\x00" * (_bit_size % 8)

    def _get_bit_values(self, key, offset, count):
        """Abstract getting bit values.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param count: The number of bits to read
        """
        key = self._get_prefix(key)
        bit_start = divmod(offset, self._bit_size)[0]
        bit_end = divmod(offset + count, self._bit_size)[0]

        request = (f"{key}:{v}" for v in range(bit_start, bit_end + 1))
        response = self.client.mget(request)
        return response

    def _val_bit(self, key, offset, count):
        """Validate that the given range is currently set in redis.

        If any of the keys return None, then it is invalid.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param count: The number of bits to read
        """
        response = self._get_bit_values(key, offset, count)
        return (
            True  # pylint: disable=simplifiable-if-expression
            if None not in response
            else False
        )

    def _get_bit(self, key, offset, count):
        """Get bit.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param count: The number of bits to read
        """
        response = self._get_bit_values(key, offset, count)
        response = (r or self._bit_default for r in response)
        result = "".join(response)
        result = unpack_bitstring(result)
        return result[offset : offset + count]

    def _set_bit(self, key, offset, values):
        """Set bit.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param values: The values to set
        """
        count = len(values)
        bit_start = divmod(offset, self._bit_size)[0]
        bit_end = divmod(offset + count, self._bit_size)[0]
        value = pack_bitstring(values)

        current = self._get_bit_values(key, offset, count)
        current = (r or self._bit_default for r in current)
        current = "".join(current)
        current = current[0:offset] + value.decode("utf-8") + current[offset + count :]
        final = (
            current[s : s + self._bit_size] for s in range(0, count, self._bit_size)
        )

        key = self._get_prefix(key)
        request = (f"{key}:{v}" for v in range(bit_start, bit_end + 1))
        request = dict(zip(request, final))
        self.client.mset(request)

    # --------------------------------------------------------------------------#
    #  Redis register implementation
    # --------------------------------------------------------------------------#
    _reg_size = 16
    _reg_default = "\x00" * (_reg_size % 8)

    def _get_reg_values(self, key, offset, count):
        """Abstract getting register values.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param count: The number of bits to read
        """
        key = self._get_prefix(key)
        request = (f"{key}:{v}" for v in range(offset, count + 1))
        response = self.client.mget(request)
        return response

    def _val_reg(self, key, offset, count):
        """Validate that the given range is currently set in redis.

        If any of the keys return None, then it is invalid.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param count: The number of bits to read
        """
        response = self._get_reg_values(key, offset, count)
        return None not in response

    def _get_reg(self, key, offset, count):
        """Get register.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param count: The number of bits to read
        """
        response = self._get_reg_values(key, offset, count)
        response = [r or self._reg_default for r in response]
        return response[offset : offset + count]

    def _set_reg(self, key, offset, values):
        """Set register.

        :param key: The key prefix to use
        :param offset: The address offset to start at
        :param values: The values to set
        """
        count = len(values)
        key = self._get_prefix(key)
        request = (f"{key}:{v}" for v in range(offset, count + 1))
        request = dict(zip(request, values))
        self.client.mset(request)



#!/usr/bin/env python3
"""Simple Asynchronous Modbus TCP over TLS client.

This is a simple example of writing a asynchronous modbus TCP over TLS client
that uses Python builtin module ssl - TLS/SSL wrapper for socket objects for
the TLS feature and asyncio.
"""
import asyncio

# -------------------------------------------------------------------------- #
# import necessary libraries
# -------------------------------------------------------------------------- #
import ssl

from pymodbus.client import AsyncModbusTlsClient


# -------------------------------------------------------------------------- #
# the TLS detail security can be set in SSLContext which is the context here
# -------------------------------------------------------------------------- #
sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
sslctx.verify_mode = ssl.CERT_REQUIRED
sslctx.check_hostname = True

# Prepare client"s certificate which the server requires for TLS full handshake
# sslctx.load_cert_chain(certfile="client.crt", keyfile="client.key",
#                        password="pwd")


async def start_async_test(client):
    """Start async test."""
    result = await client.read_coils(1, 8)
    print(result.bits)
    await client.write_coils(1, [False] * 3)
    result = await client.read_coils(1, 8)
    print(result.bits)


if __name__ == "__main__":
    # ----------------------------------------------------------------------- #
    # pass SSLContext which is the context here to ModbusTcpClient()
    # ----------------------------------------------------------------------- #
    new_client = AsyncModbusTlsClient(  # pylint: disable=unpacking-non-sequence
        "test.host.com",
        8020,
        sslctx=sslctx,
    )
    loop = asyncio.get_running_loop()
    loop.run_until_complete(start_async_test(new_client.protocol))
    loop.close()



"""Test all messages."""
import unittest

from pymodbus.bit_read_message import (
    ReadCoilsRequest,
    ReadCoilsResponse,
    ReadDiscreteInputsRequest,
    ReadDiscreteInputsResponse,
)
from pymodbus.bit_write_message import (
    WriteMultipleCoilsRequest,
    WriteMultipleCoilsResponse,
    WriteSingleCoilRequest,
    WriteSingleCoilResponse,
)
from pymodbus.constants import Defaults
from pymodbus.register_read_message import (
    ReadHoldingRegistersRequest,
    ReadHoldingRegistersResponse,
    ReadInputRegistersRequest,
    ReadInputRegistersResponse,
    ReadWriteMultipleRegistersRequest,
    ReadWriteMultipleRegistersResponse,
)
from pymodbus.register_write_message import (
    WriteMultipleRegistersRequest,
    WriteMultipleRegistersResponse,
    WriteSingleRegisterRequest,
    WriteSingleRegisterResponse,
)


# ---------------------------------------------------------------------------#
#  Fixture
# ---------------------------------------------------------------------------#


class ModbusAllMessagesTests(unittest.TestCase):
    """All messages tests."""

    # -----------------------------------------------------------------------#
    #  Setup/TearDown
    # -----------------------------------------------------------------------#

    def setUp(self):
        """Initialize the test environment and builds request/result encoding pairs."""
        arguments = {
            "read_address": 1,
            "read_count": 1,
            "write_address": 1,
            "write_registers": 1,
        }
        self.requests = [
            lambda unit: ReadCoilsRequest(1, 5, unit=unit),
            lambda unit: ReadDiscreteInputsRequest(1, 5, unit=unit),
            lambda unit: WriteSingleCoilRequest(1, 1, unit=unit),
            lambda unit: WriteMultipleCoilsRequest(1, [1], unit=unit),
            lambda unit: ReadHoldingRegistersRequest(1, 5, unit=unit),
            lambda unit: ReadInputRegistersRequest(1, 5, unit=unit),
            lambda unit: ReadWriteMultipleRegistersRequest(unit=unit, **arguments),
            lambda unit: WriteSingleRegisterRequest(1, 1, unit=unit),
            lambda unit: WriteMultipleRegistersRequest(1, [1], unit=unit),
        ]
        self.responses = [
            lambda unit: ReadCoilsResponse([1], unit=unit),
            lambda unit: ReadDiscreteInputsResponse([1], unit=unit),
            lambda unit: WriteSingleCoilResponse(1, 1, unit=unit),
            lambda unit: WriteMultipleCoilsResponse(1, [1], unit=unit),
            lambda unit: ReadHoldingRegistersResponse([1], unit=unit),
            lambda unit: ReadInputRegistersResponse([1], unit=unit),
            lambda unit: ReadWriteMultipleRegistersResponse([1], unit=unit),
            lambda unit: WriteSingleRegisterResponse(1, 1, unit=unit),
            lambda unit: WriteMultipleRegistersResponse(1, 1, unit=unit),
        ]

    def tearDown(self):
        """Clean up the test environment"""

    def test_initializing_slave_address_request(self):
        """Test that every request can initialize the unit id"""
        unit_id = 0x12
        for factory in self.requests:
            request = factory(unit_id)
            self.assertEqual(request.unit_id, unit_id)

    def test_initializing_slave_address_response(self):
        """Test that every response can initialize the unit id"""
        unit_id = 0x12
        for factory in self.responses:
            response = factory(unit_id)
            self.assertEqual(response.unit_id, unit_id)

    def test_forwarding_kwargs_to_pdu(self):
        """Test that the kwargs are forwarded to the pdu correctly"""
        request = ReadCoilsRequest(1, 5, unit=0x12, transaction=0x12, protocol=0x12)
        self.assertEqual(request.unit_id, 0x12)
        self.assertEqual(request.transaction_id, 0x12)
        self.assertEqual(request.protocol_id, 0x12)

        request = ReadCoilsRequest(1, 5)
        self.assertEqual(request.unit_id, Defaults.Slave)
        self.assertEqual(request.transaction_id, Defaults.TransactionId)
        self.assertEqual(request.protocol_id, Defaults.ProtocolId)



"""Bit Message Test Fixture.

This fixture tests the functionality of all the
bit based request/response messages:

* Read/Write Discretes
* Read Coils
"""
from test.conftest import FakeList, MockContext
import unittest

from pymodbus.bit_write_message import (
    WriteMultipleCoilsRequest,
    WriteMultipleCoilsResponse,
    WriteSingleCoilRequest,
    WriteSingleCoilResponse,
)
from pymodbus.pdu import ModbusExceptions


# ---------------------------------------------------------------------------#
#  Fixture
# ---------------------------------------------------------------------------#


class ModbusBitMessageTests(unittest.TestCase):
    """Modbus bit write message tests."""

    # -----------------------------------------------------------------------#
    #  Setup/TearDown
    # -----------------------------------------------------------------------#

    def setUp(self):
        """Initialize the test environment and builds request/result encoding pairs."""

    def tearDown(self):
        """Clean up the test environment"""

    def test_bit_write_base_requests(self):
        """Test bit write base."""
        messages = {
            WriteSingleCoilRequest(1, 0xABCD): b"\x00\x01\xff\x00",
            WriteSingleCoilResponse(1, 0xABCD): b"\x00\x01\xff\x00",
            WriteMultipleCoilsRequest(1, [True] * 5): b"\x00\x01\x00\x05\x01\x1f",
            WriteMultipleCoilsResponse(1, 5): b"\x00\x01\x00\x05",
        }
        for request, expected in iter(messages.items()):
            self.assertEqual(request.encode(), expected)

    def test_bit_write_message_get_response_pdu(self):
        """Test bit write message."""
        requests = {WriteSingleCoilRequest(1, 0xABCD): 5}
        for request, expected in iter(requests.items()):
            pdu_len = request.get_response_pdu_size()
            self.assertEqual(pdu_len, expected)

    def test_write_multiple_coils_request(self):
        """Test write multiple coils."""
        request = WriteMultipleCoilsRequest(1, [True] * 5)
        request.decode(b"\x00\x01\x00\x05\x01\x1f")
        self.assertEqual(request.byte_count, 1)
        self.assertEqual(request.address, 1)
        self.assertEqual(request.values, [True] * 5)
        self.assertEqual(request.get_response_pdu_size(), 5)

    def test_invalid_write_multiple_coils_request(self):
        """Test write invalid multiple coils."""
        request = WriteMultipleCoilsRequest(1, None)
        self.assertEqual(request.values, [])

    def test_write_single_coil_request_encode(self):
        """Test write single coil."""
        request = WriteSingleCoilRequest(1, False)
        self.assertEqual(request.encode(), b"\x00\x01\x00\x00")

    def test_write_single_coil_execute(self):
        """Test write single coil."""
        context = MockContext(False, default=True)
        request = WriteSingleCoilRequest(2, True)
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalAddress)

        context.valid = True
        result = request.execute(context)
        self.assertEqual(result.encode(), b"\x00\x02\xff\x00")

        context = MockContext(True, default=False)
        request = WriteSingleCoilRequest(2, False)
        result = request.execute(context)
        self.assertEqual(result.encode(), b"\x00\x02\x00\x00")

    def test_write_multiple_coils_execute(self):
        """Test write multiple coils."""
        context = MockContext(False)
        # too many values
        request = WriteMultipleCoilsRequest(2, FakeList(0x123456))
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalValue)

        # bad byte count
        request = WriteMultipleCoilsRequest(2, [0x00] * 4)
        request.byte_count = 0x00
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalValue)

        # does not validate
        context.valid = False
        request = WriteMultipleCoilsRequest(2, [0x00] * 4)
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalAddress)

        # validated request
        context.valid = True
        result = request.execute(context)
        self.assertEqual(result.encode(), b"\x00\x02\x00\x04")

    def test_write_multiple_coils_response(self):
        """Test write multiple coils."""
        response = WriteMultipleCoilsResponse()
        response.decode(b"\x00\x80\x00\x08")
        self.assertEqual(response.address, 0x80)
        self.assertEqual(response.count, 0x08)

    def test_serializing_to_string(self):
        """Test serializing to string."""
        requests = [
            WriteSingleCoilRequest(1, 0xABCD),
            WriteSingleCoilResponse(1, 0xABCD),
            WriteMultipleCoilsRequest(1, [True] * 5),
            WriteMultipleCoilsResponse(1, 5),
        ]
        for request in requests:
            result = str(request)
            self.assertTrue(result is not None and len(result))



"""Test register write messages."""
from test.conftest import MockContext, MockLastValuesContext
import unittest

from pymodbus.payload import BinaryPayloadBuilder, Endian
from pymodbus.pdu import ModbusExceptions
from pymodbus.register_write_message import (
    MaskWriteRegisterRequest,
    MaskWriteRegisterResponse,
    WriteMultipleRegistersRequest,
    WriteMultipleRegistersResponse,
    WriteSingleRegisterRequest,
    WriteSingleRegisterResponse,
)


# ---------------------------------------------------------------------------#
#  Fixture
# ---------------------------------------------------------------------------#


class WriteRegisterMessagesTest(unittest.TestCase):
    """Register Message Test Fixture.

    This fixture tests the functionality of all the
    register based request/response messages:

    * Read/Write Input Registers
    * Read Holding Registers
    """

    def setUp(self):
        """Initialize the test environment and builds request/result encoding pairs."""
        self.value = 0xABCD
        self.values = [0xA, 0xB, 0xC]
        builder = BinaryPayloadBuilder(byteorder=Endian.Big)
        builder.add_16bit_uint(0x1234)
        self.payload = builder.build()
        self.write = {
            WriteSingleRegisterRequest(1, self.value): b"\x00\x01\xab\xcd",
            WriteSingleRegisterResponse(1, self.value): b"\x00\x01\xab\xcd",
            WriteMultipleRegistersRequest(
                1, self.values
            ): b"\x00\x01\x00\x03\x06\x00\n\x00\x0b\x00\x0c",
            WriteMultipleRegistersResponse(1, 5): b"\x00\x01\x00\x05",
            WriteSingleRegisterRequest(
                1, self.payload[0], skip_encode=True
            ): b"\x00\x01\x12\x34",
            WriteMultipleRegistersRequest(
                1, self.payload, skip_encode=True
            ): b"\x00\x01\x00\x01\x02\x12\x34",
        }

    def tearDown(self):
        """Clean up the test environment"""
        del self.write

    def test_register_write_requests_encode(self):
        """Test register write requests encode."""
        for request, response in iter(self.write.items()):
            self.assertEqual(request.encode(), response)

    def test_register_write_requests_decode(self):
        """Test register write requests decode."""
        addresses = [1, 1, 1, 1]
        values = sorted(
            self.write.items(),
            key=lambda x: str(x),  # pylint: disable=unnecessary-lambda
        )
        for packet, address in zip(values, addresses):
            request, response = packet
            request.decode(response)
            self.assertEqual(request.address, address)

    def test_invalid_write_multiple_registers_request(self):
        """Test invalid write multiple registers request."""
        request = WriteMultipleRegistersRequest(0, None)
        self.assertEqual(request.values, [])

    def test_serializing_to_string(self):
        """Test serializing to string."""
        for request in iter(self.write.keys()):
            self.assertTrue(str(request) is not None)

    def test_write_single_register_request(self):
        """Test write single register request."""
        context = MockContext()
        request = WriteSingleRegisterRequest(0x00, 0xF0000)
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalValue)

        request.value = 0x00FF
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalAddress)

        context.valid = True
        result = request.execute(context)
        self.assertEqual(result.function_code, request.function_code)

    def test_write_multiple_register_request(self):
        """Test write multiple register request."""
        context = MockContext()
        request = WriteMultipleRegistersRequest(0x00, [0x00] * 10)
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalAddress)

        request.count = 0x05  # bytecode != code * 2
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalValue)

        request.count = 0x800  # outside of range
        result = request.execute(context)
        self.assertEqual(result.exception_code, ModbusExceptions.IllegalValue)

        context.valid = True
        request = WriteMultipleRegistersRequest(0x00, [0x00] * 10)
        result = request.execute(context)
        self.assertEqual(result.function_code, request.function_code)

        # -----------------------------------------------------------------------#
        # Mask Write Register Request
        # -----------------------------------------------------------------------#

    def test_mask_write_register_request_encode(self):
        """Test basic bit message encoding/decoding"""
        handle = MaskWriteRegisterRequest(0x0000, 0x0101, 0x1010)
        result = handle.encode()
        self.assertEqual(result, b"\x00\x00\x01\x01\x10\x10")

    def test_mask_write_register_request_decode(self):
        """Test basic bit message encoding/decoding"""
        request = b"\x00\x04\x00\xf2\x00\x25"
        handle = MaskWriteRegisterRequest()
        handle.decode(request)
        self.assertEqual(handle.address, 0x0004)
        self.assertEqual(handle.and_mask, 0x00F2)
        self.assertEqual(handle.or_mask, 0x0025)

    def test_mask_write_register_request_execute(self):
        """Test write register request valid execution"""
        # The test uses the 4 nibbles of the 16-bit values to test
        # the combinations:
        #     and_mask=0, or_mask=0
        #     and_mask=F, or_mask=0
        #     and_mask=0, or_mask=F
        #     and_mask=F, or_mask=F
        context = MockLastValuesContext(valid=True, default=0xAA55)
        handle = MaskWriteRegisterRequest(0x0000, 0x0F0F, 0x00FF)
        result = handle.execute(context)
        self.assertTrue(isinstance(result, MaskWriteRegisterResponse))
        self.assertEqual([0x0AF5], context.last_values)

    def test_mask_write_register_request_invalid_execute(self):
        """Test write register request execute with invalid data"""
        context = MockContext(valid=False, default=0x0000)
        handle = MaskWriteRegisterRequest(0x0000, -1, 0x1010)
        result = handle.execute(context)
        self.assertEqual(ModbusExceptions.IllegalValue, result.exception_code)

        handle = MaskWriteRegisterRequest(0x0000, 0x0101, -1)
        result = handle.execute(context)
        self.assertEqual(ModbusExceptions.IllegalValue, result.exception_code)

        handle = MaskWriteRegisterRequest(0x0000, 0x0101, 0x1010)
        result = handle.execute(context)
        self.assertEqual(ModbusExceptions.IllegalAddress, result.exception_code)

        # -----------------------------------------------------------------------#
        # Mask Write Register Response
        # -----------------------------------------------------------------------#

    def test_mask_write_register_response_encode(self):
        """Test basic bit message encoding/decoding"""
        handle = MaskWriteRegisterResponse(0x0000, 0x0101, 0x1010)
        result = handle.encode()
        self.assertEqual(result, b"\x00\x00\x01\x01\x10\x10")

    def test_mask_write_register_response_decode(self):
        """Test basic bit message encoding/decoding"""
        request = b"\x00\x04\x00\xf2\x00\x25"
        handle = MaskWriteRegisterResponse()
        handle.decode(request)
        self.assertEqual(handle.address, 0x0004)
        self.assertEqual(handle.and_mask, 0x00F2)
        self.assertEqual(handle.or_mask, 0x0025)



"""Bit Message Test Fixture.

This fixture tests the functionality of all the
bit based request/response messages:

* Read/Write Discretes
* Read Coils
"""
import struct

from test.conftest import MockContext
import unittest

from pymodbus.bit_read_message import (
    ReadBitsRequestBase,
    ReadBitsResponseBase,
    ReadCoilsRequest,
    ReadDiscreteInputsRequest,
)
from pymodbus.pdu import ModbusExceptions


res = [True] * 21
res.extend([False] * 3)
# ---------------------------------------------------------------------------#
#  Fixture
# ---------------------------------------------------------------------------#


class ModbusBitMessageTests(unittest.TestCase):
    """Modbus bit read message tests."""

    # -----------------------------------------------------------------------#
    #  Setup/TearDown
    # -----------------------------------------------------------------------#

    def setUp(self):
        """Initialize the test environment and builds request/result encoding pairs."""

    def tearDown(self):
        """Clean up the test environment"""

    def test_read_bit_base_class_methods(self):
        """Test basic bit message encoding/decoding"""
        handle = ReadBitsRequestBase(1, 1)
        msg = "ReadBitRequest(1,1)"
        self.assertEqual(msg, str(handle))
        handle = ReadBitsResponseBase([1, 1])
        msg = "ReadBitsResponseBase(2)"
        self.assertEqual(msg, str(handle))

    def test_bit_read_base_request_encoding(self):
        """Test basic bit message encoding/decoding"""
        for i in range(20):
            handle = ReadBitsRequestBase(i, i)
            result = struct.pack(">HH", i, i)
            self.assertEqual(handle.encode(), result)
            handle.decode(result)
            self.assertEqual((handle.address, handle.count), (i, i))

    def test_bit_read_base_response_encoding(self):
        """Test basic bit message encoding/decoding"""
        for i in range(20):
            data = [True] * i
            handle = ReadBitsResponseBase(data)
            result = handle.encode()
            handle.decode(result)
            self.assertEqual(handle.bits[:i], data)

    def test_bit_read_base_response_helper_methods(self):
        """Test the extra methods on a ReadBitsResponseBase"""
        data = [False] * 8
        handle = ReadBitsResponseBase(data)
        for i in (1, 3, 5):
            handle.setBit(i, True)
        for i in (1, 3, 5):
            handle.resetBit(i)
        for i in range(8):
            self.assertEqual(handle.getBit(i), False)

    def test_bit_read_base_requests(self):
        """Test bit read request encoding"""
        messages = {
            ReadBitsRequestBase(12, 14): b"\x00\x0c\x00\x0e",
            ReadBitsResponseBase([1, 0, 1, 1, 0]): b"\x01\x0d",
        }
        for request, expected in iter(messages.items()):
            self.assertEqual(request.encode(), expected)

    def test_bit_read_message_execute_value_errors(self):
        """Test bit read request encoding"""
        context = MockContext()
        requests = [
            ReadCoilsRequest(1, 0x800),
            ReadDiscreteInputsRequest(1, 0x800),
        ]
        for request in requests:
            result = request.execute(context)
            self.assertEqual(ModbusExceptions.IllegalValue, result.exception_code)

    def test_bit_read_message_execute_address_errors(self):
        """Test bit read request encoding"""
        context = MockContext()
        requests = [
            ReadCoilsRequest(1, 5),
            ReadDiscreteInputsRequest(1, 5),
        ]
        for request in requests:
            result = request.execute(context)
            self.assertEqual(ModbusExceptions.IllegalAddress, result.exception_code)

    def test_bit_read_message_execute_success(self):
        """Test bit read request encoding"""
        context = MockContext()
        context.validate = lambda a, b, c: True
        requests = [
            ReadCoilsRequest(1, 5),
            ReadDiscreteInputsRequest(1, 5),
        ]
        for request in requests:
            result = request.execute(context)
            self.assertEqual(result.bits, [True] * 5)

    def test_bit_read_message_get_response_pdu(self):
        """Test bit read message get response pdu."""
        requests = {
            ReadCoilsRequest(1, 5): 3,
            ReadCoilsRequest(1, 8): 3,
            ReadCoilsRequest(0, 16): 4,
            ReadDiscreteInputsRequest(1, 21): 5,
            ReadDiscreteInputsRequest(1, 24): 5,
            ReadDiscreteInputsRequest(1, 1900): 240,
        }
        for request, expected in iter(requests.items()):
            pdu_len = request.get_response_pdu_size()
            self.assertEqual(pdu_len, expected)



"""Test diag messages."""
import unittest

from pymodbus.constants import ModbusPlusOperation
from pymodbus.diag_message import (
    ChangeAsciiInputDelimiterRequest,
    ChangeAsciiInputDelimiterResponse,
    ClearCountersRequest,
    ClearCountersResponse,
    ClearOverrunCountRequest,
    ClearOverrunCountResponse,
    DiagnosticStatusRequest,
    DiagnosticStatusResponse,
    DiagnosticStatusSimpleRequest,
    DiagnosticStatusSimpleResponse,
    ForceListenOnlyModeRequest,
    ForceListenOnlyModeResponse,
    GetClearModbusPlusRequest,
    GetClearModbusPlusResponse,
    RestartCommunicationsOptionRequest,
    RestartCommunicationsOptionResponse,
    ReturnBusCommunicationErrorCountRequest,
    ReturnBusCommunicationErrorCountResponse,
    ReturnBusExceptionErrorCountRequest,
    ReturnBusExceptionErrorCountResponse,
    ReturnBusMessageCountRequest,
    ReturnBusMessageCountResponse,
    ReturnDiagnosticRegisterRequest,
    ReturnDiagnosticRegisterResponse,
    ReturnIopOverrunCountRequest,
    ReturnIopOverrunCountResponse,
    ReturnQueryDataRequest,
    ReturnQueryDataResponse,
    ReturnSlaveBusCharacterOverrunCountRequest,
    ReturnSlaveBusCharacterOverrunCountResponse,
    ReturnSlaveBusyCountRequest,
    ReturnSlaveBusyCountResponse,
    ReturnSlaveMessageCountRequest,
    ReturnSlaveMessageCountResponse,
    ReturnSlaveNAKCountRequest,
    ReturnSlaveNAKCountResponse,
    ReturnSlaveNoReponseCountResponse,
    ReturnSlaveNoResponseCountRequest,
)
from pymodbus.exceptions import NotImplementedException


class SimpleDataStoreTest(unittest.TestCase):
    """Unittest for the pymodbus.diag_message module."""

    def setUp(self):
        """Do setup."""
        self.requests = [
            (
                RestartCommunicationsOptionRequest,
                b"\x00\x01\x00\x00",
                b"\x00\x01\xff\x00",
            ),
            (ReturnDiagnosticRegisterRequest, b"\x00\x02\x00\x00", b"\x00\x02\x00\x00"),
            (
                ChangeAsciiInputDelimiterRequest,
                b"\x00\x03\x00\x00",
                b"\x00\x03\x00\x00",
            ),
            (ForceListenOnlyModeRequest, b"\x00\x04\x00\x00", b"\x00\x04"),
            (ReturnQueryDataRequest, b"\x00\x00\x00\x00", b"\x00\x00\x00\x00"),
            (ClearCountersRequest, b"\x00\x0a\x00\x00", b"\x00\x0a\x00\x00"),
            (ReturnBusMessageCountRequest, b"\x00\x0b\x00\x00", b"\x00\x0b\x00\x00"),
            (
                ReturnBusCommunicationErrorCountRequest,
                b"\x00\x0c\x00\x00",
                b"\x00\x0c\x00\x00",
            ),
            (
                ReturnBusExceptionErrorCountRequest,
                b"\x00\x0d\x00\x00",
                b"\x00\x0d\x00\x00",
            ),
            (ReturnSlaveMessageCountRequest, b"\x00\x0e\x00\x00", b"\x00\x0e\x00\x00"),
            (
                ReturnSlaveNoResponseCountRequest,
                b"\x00\x0f\x00\x00",
                b"\x00\x0f\x00\x00",
            ),
            (ReturnSlaveNAKCountRequest, b"\x00\x10\x00\x00", b"\x00\x10\x00\x00"),
            (ReturnSlaveBusyCountRequest, b"\x00\x11\x00\x00", b"\x00\x11\x00\x00"),
            (
                ReturnSlaveBusCharacterOverrunCountRequest,
                b"\x00\x12\x00\x00",
                b"\x00\x12\x00\x00",
            ),
            (ReturnIopOverrunCountRequest, b"\x00\x13\x00\x00", b"\x00\x13\x00\x00"),
            (ClearOverrunCountRequest, b"\x00\x14\x00\x00", b"\x00\x14\x00\x00"),
            (
                GetClearModbusPlusRequest,
                b"\x00\x15\x00\x00",
                b"\x00\x15\x00\x00" + b"\x00\x00" * 55,
            ),
        ]

        self.responses = [
            # (DiagnosticStatusResponse,                     b"\x00\x00\x00\x00"),
            # (DiagnosticStatusSimpleResponse,               b"\x00\x00\x00\x00"),
            (ReturnQueryDataResponse, b"\x00\x00\x00\x00"),
            (RestartCommunicationsOptionResponse, b"\x00\x01\x00\x00"),
            (ReturnDiagnosticRegisterResponse, b"\x00\x02\x00\x00"),
            (ChangeAsciiInputDelimiterResponse, b"\x00\x03\x00\x00"),
            (ForceListenOnlyModeResponse, b"\x00\x04"),
            (ReturnQueryDataResponse, b"\x00\x00\x00\x00"),
            (ClearCountersResponse, b"\x00\x0a\x00\x00"),
            (ReturnBusMessageCountResponse, b"\x00\x0b\x00\x00"),
            (ReturnBusCommunicationErrorCountResponse, b"\x00\x0c\x00\x00"),
            (ReturnBusExceptionErrorCountResponse, b"\x00\x0d\x00\x00"),
            (ReturnSlaveMessageCountResponse, b"\x00\x0e\x00\x00"),
            (ReturnSlaveNoReponseCountResponse, b"\x00\x0f\x00\x00"),
            (ReturnSlaveNAKCountResponse, b"\x00\x10\x00\x00"),
            (ReturnSlaveBusyCountResponse, b"\x00\x11\x00\x00"),
            (ReturnSlaveBusCharacterOverrunCountResponse, b"\x00\x12\x00\x00"),
            (ReturnIopOverrunCountResponse, b"\x00\x13\x00\x00"),
            (ClearOverrunCountResponse, b"\x00\x14\x00\x00"),
            (GetClearModbusPlusResponse, b"\x00\x15\x00\x04" + b"\x00\x00" * 55),
        ]

    def tearDown(self):
        """Clean up the test environment"""
        del self.requests
        del self.responses

    def test_diagnostic_requests_decode(self):
        """Testing diagnostic request messages encoding"""
        for msg, enc, _ in self.requests:
            handle = DiagnosticStatusRequest()
            handle.decode(enc)
            self.assertEqual(handle.sub_function_code, msg.sub_function_code)

    def test_diagnostic_simple_requests(self):
        """Testing diagnostic request messages encoding"""
        request = DiagnosticStatusSimpleRequest(b"\x12\x34")
        request.sub_function_code = 0x1234
        self.assertRaises(
            NotImplementedException,
            lambda: request.execute(),  # pylint: disable=unnecessary-lambda
        )
        self.assertEqual(request.encode(), b"\x12\x34\x12\x34")
        DiagnosticStatusSimpleResponse(None)

    def test_diagnostic_response_decode(self):
        """Testing diagnostic request messages encoding"""
        for msg, enc, _ in self.requests:
            handle = DiagnosticStatusResponse()
            handle.decode(enc)
            self.assertEqual(handle.sub_function_code, msg.sub_function_code)

    def test_diagnostic_requests_encode(self):
        """Testing diagnostic request messages encoding"""
        for msg, enc, _ in self.requests:
            self.assertEqual(msg().encode(), enc)

    def test_diagnostic_execute(self):
        """Testing diagnostic message execution"""
        for message, encoded, executed in self.requests:
            encoded = message().execute().encode()
            self.assertEqual(encoded, executed)

    def test_return_query_data_request(self):
        """Testing diagnostic message execution"""
        message = ReturnQueryDataRequest([0x0000] * 2)
        self.assertEqual(message.encode(), b"\x00\x00\x00\x00\x00\x00")
        message = ReturnQueryDataRequest(0x0000)
        self.assertEqual(message.encode(), b"\x00\x00\x00\x00")

    def test_return_query_data_response(self):
        """Testing diagnostic message execution"""
        message = ReturnQueryDataResponse([0x0000] * 2)
        self.assertEqual(message.encode(), b"\x00\x00\x00\x00\x00\x00")
        message = ReturnQueryDataResponse(0x0000)
        self.assertEqual(message.encode(), b"\x00\x00\x00\x00")

    def test_restart_cmmunications_option(self):
        """Testing diagnostic message execution"""
        request = RestartCommunicationsOptionRequest(True)
        self.assertEqual(request.encode(), b"\x00\x01\xff\x00")
        request = RestartCommunicationsOptionRequest(False)
        self.assertEqual(request.encode(), b"\x00\x01\x00\x00")

        response = RestartCommunicationsOptionResponse(True)
        self.assertEqual(response.encode(), b"\x00\x01\xff\x00")
        response = RestartCommunicationsOptionResponse(False)
        self.assertEqual(response.encode(), b"\x00\x01\x00\x00")

    def test_get_clear_modbus_plus_request_execute(self):
        """Testing diagnostic message execution"""
        request = GetClearModbusPlusRequest(data=ModbusPlusOperation.ClearStatistics)
        response = request.execute()
        self.assertEqual(response.message, ModbusPlusOperation.ClearStatistics)

        request = GetClearModbusPlusRequest(data=ModbusPlusOperation.GetStatistics)
        response = request.execute()
        resp = [ModbusPlusOperation.GetStatistics]
        self.assertEqual(response.message, resp + [0x00] * 55)



"""Modbus Payload Builders.

A collection of utilities for building and decoding
modbus messages payloads.
"""
# pylint: disable=missing-type-doc
import logging
from struct import pack, unpack

from pymodbus.constants import Endian
from pymodbus.exceptions import ParameterException
from pymodbus.interfaces import IPayloadBuilder
from pymodbus.utilities import (
    make_byte_string,
    pack_bitstring,
    unpack_bitstring,
)


# --------------------------------------------------------------------------- #
# Logging
# --------------------------------------------------------------------------- #
_logger = logging.getLogger(__name__)


WC = {"b": 1, "h": 2, "e": 2, "i": 4, "l": 4, "q": 8, "f": 4, "d": 8}


class BinaryPayloadBuilder(IPayloadBuilder):
    """A utility that helps build payload messages to be written with the various modbus messages.

    It really is just a simple wrapper around the struct module,
    however it saves time looking up the format strings.
    What follows is a simple example::

        builder = BinaryPayloadBuilder(byteorder=Endian.Little)
        builder.add_8bit_uint(1)
        builder.add_16bit_uint(2)
        payload = builder.build()
    """

    def __init__(
        self, payload=None, byteorder=Endian.Little, wordorder=Endian.Big, repack=False
    ):
        """Initialize a new instance of the payload builder.

        :param payload: Raw binary payload data to initialize with
        :param byteorder: The endianness of the bytes in the words
        :param wordorder: The endianness of the word (when wordcount is >= 2)
        :param repack: Repack the provided payload based on BO
        """
        self._payload = payload or []
        self._byteorder = byteorder
        self._wordorder = wordorder
        self._repack = repack

    def _pack_words(self, fstring, value):
        """Pack words based on the word order and byte order.

        # ---------------------------------------------- #
        # pack in to network ordered value               #
        # unpack in to network ordered  unsigned integer #
        # Change Word order if little endian word order  #
        # Pack values back based on correct byte order   #
        # ---------------------------------------------- #

        :param fstring:
        :param value: Value to be packed
        :return:
        """
        value = pack(
            "!{}".format(fstring), value  # pylint: disable=consider-using-f-string
        )
        wordorder = WC.get(fstring.lower()) // 2
        upperbyte = f"!{wordorder}H"
        payload = unpack(upperbyte, value)

        if self._wordorder == Endian.Little:
            payload = list(reversed(payload))

        fstring = self._byteorder + "H"
        payload = [pack(fstring, word) for word in payload]
        payload = b"".join(payload)

        return payload

    def to_string(self):
        """Return the payload buffer as a string.

        :returns: The payload buffer as a string
        """
        return b"".join(self._payload)

    def __str__(self):
        """Return the payload buffer as a string.

        :returns: The payload buffer as a string
        """
        return self.to_string().decode("utf-8")

    def reset(self):
        """Reset the payload buffer."""
        self._payload = []

    def to_registers(self):
        """Convert the payload buffer to register layout that can be used as a context block.

        :returns: The register layout to use as a block
        """
        # fstring = self._byteorder+"H"
        fstring = "!H"
        payload = self.build()
        if self._repack:
            payload = [unpack(self._byteorder + "H", value)[0] for value in payload]
        else:
            payload = [unpack(fstring, value)[0] for value in payload]
        _logger.debug(payload)
        return payload

    def to_coils(self):
        """Convert the payload buffer into a coil layout that can be used as a context block.

        :returns: The coil layout to use as a block
        """
        payload = self.to_registers()
        coils = [bool(int(bit)) for reg in payload for bit in format(reg, "016b")]
        return coils

    def build(self):
        """Return the payload buffer as a list.

        This list is two bytes per element and can
        thus be treated as a list of registers.

        :returns: The payload buffer as a list
        """
        string = self.to_string()
        length = len(string)
        string = string + (b"\x00" * (length % 2))
        return [string[i : i + 2] for i in range(0, length, 2)]

    def add_bits(self, values):
        """Add a collection of bits to be encoded.

        If these are less than a multiple of eight,
        they will be left padded with 0 bits to make
        it so.

        :param values: The value to add to the buffer
        """
        value = pack_bitstring(values)
        self._payload.append(value)

    def add_8bit_uint(self, value):
        """Add a 8 bit unsigned int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = self._byteorder + "B"
        self._payload.append(pack(fstring, value))

    def add_16bit_uint(self, value):
        """Add a 16 bit unsigned int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = self._byteorder + "H"
        self._payload.append(pack(fstring, value))

    def add_32bit_uint(self, value):
        """Add a 32 bit unsigned int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = "I"
        # fstring = self._byteorder + "I"
        p_string = self._pack_words(fstring, value)
        self._payload.append(p_string)

    def add_64bit_uint(self, value):
        """Add a 64 bit unsigned int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = "Q"
        p_string = self._pack_words(fstring, value)
        self._payload.append(p_string)

    def add_8bit_int(self, value):
        """Add a 8 bit signed int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = self._byteorder + "b"
        self._payload.append(pack(fstring, value))

    def add_16bit_int(self, value):
        """Add a 16 bit signed int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = self._byteorder + "h"
        self._payload.append(pack(fstring, value))

    def add_32bit_int(self, value):
        """Add a 32 bit signed int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = "i"
        p_string = self._pack_words(fstring, value)
        self._payload.append(p_string)

    def add_64bit_int(self, value):
        """Add a 64 bit signed int to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = "q"
        p_string = self._pack_words(fstring, value)
        self._payload.append(p_string)

    def add_16bit_float(self, value):
        """Add a 16 bit float to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = "e"
        p_string = self._pack_words(fstring, value)
        self._payload.append(p_string)

    def add_32bit_float(self, value):
        """Add a 32 bit float to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = "f"
        p_string = self._pack_words(fstring, value)
        self._payload.append(p_string)

    def add_64bit_float(self, value):
        """Add a 64 bit float(double) to the buffer.

        :param value: The value to add to the buffer
        """
        fstring = "d"
        p_string = self._pack_words(fstring, value)
        self._payload.append(p_string)

    def add_string(self, value):
        """Add a string to the buffer.

        :param value: The value to add to the buffer
        """
        value = make_byte_string(value)
        fstring = self._byteorder + str(len(value)) + "s"
        self._payload.append(pack(fstring, value))


class BinaryPayloadDecoder:
    """A utility that helps decode payload messages from a modbus response message.

    It really is just a simple wrapper around
    the struct module, however it saves time looking up the format
    strings. What follows is a simple example::

        decoder = BinaryPayloadDecoder(payload)
        first   = decoder.decode_8bit_uint()
        second  = decoder.decode_16bit_uint()
    """

    def __init__(self, payload, byteorder=Endian.Little, wordorder=Endian.Big):
        """Initialize a new payload decoder.

        :param payload: The payload to decode with
        :param byteorder: The endianness of the payload
        :param wordorder: The endianness of the word (when wordcount is >= 2)
        """
        self._payload = payload
        self._pointer = 0x00
        self._byteorder = byteorder
        self._wordorder = wordorder

    @classmethod
    def fromRegisters(  # pylint: disable=invalid-name
        cls,
        registers,
        byteorder=Endian.Little,
        wordorder=Endian.Big,
    ):
        """Initialize a payload decoder.

        With the result of reading a collection of registers from a modbus device.

        The registers are treated as a list of 2 byte values.
        We have to do this because of how the data has already
        been decoded by the rest of the library.

        :param registers: The register results to initialize with
        :param byteorder: The Byte order of each word
        :param wordorder: The endianness of the word (when wordcount is >= 2)
        :returns: An initialized PayloadDecoder
        :raises ParameterException:
        """
        _logger.debug(registers)
        if isinstance(registers, list):  # repack into flat binary
            payload = b"".join(pack("!H", x) for x in registers)
            return cls(payload, byteorder, wordorder)
        raise ParameterException("Invalid collection of registers supplied")

    @classmethod
    def bit_chunks(cls, coils, size=8):
        """Return bit chunks."""
        chunks = [coils[i : i + size] for i in range(0, len(coils), size)]
        return chunks

    @classmethod
    def fromCoils(  # pylint: disable=invalid-name
        cls,
        coils,
        byteorder=Endian.Little,
        wordorder=Endian.Big,
    ):  # pylint: disable=unused-argument
        """Initialize a payload decoder with the result of reading of coils.

        The coils are treated as a list of bit(boolean) values.

        :param coils: The coil results to initialize with
        :param byteorder: The endianness of the payload
        :param wordorder: The endianness of the payload
        :returns: An initialized PayloadDecoder
        :raises ParameterException:
        """
        if isinstance(coils, list):
            payload = b""
            if padding := len(coils) % 8:  # Pad zero"s
                extra = [False] * padding
                coils = extra + coils
            chunks = cls.bit_chunks(coils)
            for chunk in chunks:
                payload += pack_bitstring(chunk[::-1])
            return cls(payload, byteorder)
        raise ParameterException("Invalid collection of coils supplied")

    def _unpack_words(self, fstring, handle):
        """Unpack words based on the word order and byte order.

        # ---------------------------------------------- #
        # Unpack in to network ordered unsigned integer  #
        # Change Word order if little endian word order  #
        # Pack values back based on correct byte order   #
        # ---------------------------------------------- #
        :param fstring:
        :param handle: Value to be unpacked
        :return:
        """
        handle = make_byte_string(handle)
        wc_value = WC.get(fstring.lower()) // 2
        handle = unpack(f"!{wc_value}H", handle)
        if self._wordorder == Endian.Little:
            handle = list(reversed(handle))

        # Repack as unsigned Integer
        handle = [pack(self._byteorder + "H", p) for p in handle]
        _logger.debug(handle)
        handle = b"".join(handle)
        return handle

    def reset(self):
        """Reset the decoder pointer back to the start."""
        self._pointer = 0x00

    def decode_8bit_uint(self):
        """Decode a 8 bit unsigned int from the buffer."""
        self._pointer += 1
        fstring = self._byteorder + "B"
        handle = self._payload[self._pointer - 1 : self._pointer]
        handle = make_byte_string(handle)
        return unpack(fstring, handle)[0]

    def decode_bits(self, package_len=1):
        """Decode a byte worth of bits from the buffer."""
        self._pointer += package_len
        # fstring = self._endian + "B"
        handle = self._payload[self._pointer - 1 : self._pointer]
        handle = make_byte_string(handle)
        return unpack_bitstring(handle)

    def decode_16bit_uint(self):
        """Decode a 16 bit unsigned int from the buffer."""
        self._pointer += 2
        fstring = self._byteorder + "H"
        handle = self._payload[self._pointer - 2 : self._pointer]
        handle = make_byte_string(handle)
        return unpack(fstring, handle)[0]

    def decode_32bit_uint(self):
        """Decode a 32 bit unsigned int from the buffer."""
        self._pointer += 4
        fstring = "I"
        # fstring = "I"
        handle = self._payload[self._pointer - 4 : self._pointer]
        handle = self._unpack_words(fstring, handle)
        return unpack("!" + fstring, handle)[0]

    def decode_64bit_uint(self):
        """Decode a 64 bit unsigned int from the buffer."""
        self._pointer += 8
        fstring = "Q"
        handle = self._payload[self._pointer - 8 : self._pointer]
        handle = self._unpack_words(fstring, handle)
        return unpack("!" + fstring, handle)[0]

    def decode_8bit_int(self):
        """Decode a 8 bit signed int from the buffer."""
        self._pointer += 1
        fstring = self._byteorder + "b"
        handle = self._payload[self._pointer - 1 : self._pointer]
        handle = make_byte_string(handle)
        return unpack(fstring, handle)[0]

    def decode_16bit_int(self):
        """Decode a 16 bit signed int from the buffer."""
        self._pointer += 2
        fstring = self._byteorder + "h"
        handle = self._payload[self._pointer - 2 : self._pointer]
        handle = make_byte_string(handle)
        return unpack(fstring, handle)[0]

    def decode_32bit_int(self):
        """Decode a 32 bit signed int from the buffer."""
        self._pointer += 4
        fstring = "i"
        handle = self._payload[self._pointer - 4 : self._pointer]
        handle = self._unpack_words(fstring, handle)
        return unpack("!" + fstring, handle)[0]

    def decode_64bit_int(self):
        """Decode a 64 bit signed int from the buffer."""
        self._pointer += 8
        fstring = "q"
        handle = self._payload[self._pointer - 8 : self._pointer]
        handle = self._unpack_words(fstring, handle)
        return unpack("!" + fstring, handle)[0]

    def decode_16bit_float(self):
        """Decode a 16 bit float from the buffer."""
        self._pointer += 2
        fstring = "e"
        handle = self._payload[self._pointer - 2 : self._pointer]
        handle = self._unpack_words(fstring, handle)
        return unpack("!" + fstring, handle)[0]

    def decode_32bit_float(self):
        """Decode a 32 bit float from the buffer."""
        self._pointer += 4
        fstring = "f"
        handle = self._payload[self._pointer - 4 : self._pointer]
        handle = self._unpack_words(fstring, handle)
        return unpack("!" + fstring, handle)[0]

    def decode_64bit_float(self):
        """Decode a 64 bit float(double) from the buffer."""
        self._pointer += 8
        fstring = "d"
        handle = self._payload[self._pointer - 8 : self._pointer]
        handle = self._unpack_words(fstring, handle)
        return unpack("!" + fstring, handle)[0]

    def decode_string(self, size=1):
        """Decode a string from the buffer.

        :param size: The size of the string to decode
        """
        self._pointer += size
        return self._payload[self._pointer - size : self._pointer]

    def skip_bytes(self, nbytes):
        """Skip n bytes in the buffer.

        :param nbytes: The number of bytes to skip
        """
        self._pointer += nbytes


# ---------------------------------------------------------------------------#
#  Exported Identifiers
# ---------------------------------------------------------------------------#
__all__ = ["BinaryPayloadBuilder", "BinaryPayloadDecoder"]
